from ragcar.tools.tokenization import RagcarTokenizationFactory
from ragcar.tools.sentence_embedding import RagcarSentenceFactory
from ragcar.tools.semantic_textual_similarity import RagcarStsFactory
from ragcar.tools.semantic_search import RagcarSemanticSearchFactory
from ragcar.tools.text_generation import RagcarTextGenerationFactory
from ragcar.tools.text_segmentation import RagcarSegmentationFactory


import logging
from typing import Optional, Union, List

import torch
import asyncio
from sentence_transformers import util

from ragcar.tools.utils.base import RagcarFactoryBase, RagcarBiencoderBase, RagcarAsyncBiencoderBase
from ragcar.tools.utils.model_config import get_es_config


logger = logging.getLogger(__name__)


class RagcarSemanticSearchFactory(RagcarFactoryBase):
    """
    Semantic search using Elasticsearch or Bi-Encoder, returns top relevant matches.
    
    Examples:
        >>> retriever = Ragcar(tool="semantic_search", src="model_name_or_path")
        >>> retriever(
        ... ["`예술인 기회소득` 청신호… `경기국제공항` 난기류", "한국vs크로아티아 경기 주심은 왜 눈물을 닦았나[VNL]"], 
        ... ['고준호 경기도의원 "파주 5000번 버스 경기 공공버스로 전환"', "강호 독일 상대로 8경기 만에 마침내 1세트 따낸 女배구"]
        ... )
        [
            [
                {'corpus_id': 0, 'score': 0.2568105459213257}, 
                {'corpus_id': 1, 'score': 0.07774988561868668}
            ], 
            [
                {'corpus_id': 1, 'score': 0.22476400434970856}, 
                {'corpus_id': 0, 'score': 0.16569660604000092}
            ]
        ]
    """
    
    def __init__(
        self, 
        tool: str, 
        src: str, 
        model: Optional[Union[str, dict]],
        device: Optional[str] = None,
        use_async: bool = False
    ):
        super().__init__(tool, src, model, device)
        self.use_async = use_async
    
    @staticmethod
    def get_available_srcs() -> List[str]:
        return [
            "model_name_or_path",
            "googledrive",
            "openai",
            "clova",
            "elasticsearch",
        ]
    
    @staticmethod
    def get_available_models() -> dict:
        return {
            "model_name_or_path": [
                "leewaay/kpf-bert-base-klueNLI-klueSTS-MSL512",
                "leewaay/klue-roberta-base-klueNLI-klueSTS-MSL512",
                "leewaay/klue-roberta-large-klueNLI-klueSTS-MSL512",
                "MODELS_SUPPORTED(https://huggingface.co/models?pipeline_tag=sentence-similarity)"
            ],
            "googledrive": [
                "YOUR_MODEL"
            ],
            "openai": [
                "text-embedding-3-large",
                "text-embedding-3-small",
                "text-embedding-ada-002",
                "MODELS_SUPPORTED(https://platform.openai.com/docs/models)"
            ],
            "clova": [
                "YOUR_MODEL(https://www.ncloud.com/product/aiService/clovaStudio)"
            ],
            "elasticsearch": [
                "YOUR_MODEL"
            ],
        }
    
    def load(self, device: str) -> object:
        """
        Loads and initializes the model based on configuration settings.

        Args:
            device (str): The computing device ('cpu' or 'cuda') where the model will be loaded.

        Returns:
            object: The initialized model ready for use.
        """
        from ragcar.tools.sentence_embedding import RagcarSentenceFactory
        
        if self.config.src == "elasticsearch":
            es_conf = get_es_config(self.config.model_info)
            
            encoder = RagcarSentenceFactory(
                tool="sentence_embedding",
                src=es_conf.encoder_key.src,
                model=es_conf.encoder_key.model
            ).load(device)
            
            if self.use_async:
                from elasticsearch import AsyncElasticsearch
            
                es = AsyncElasticsearch(
                    es_conf.host_n, 
                    http_auth=es_conf.http_auth, 
                    scheme=es_conf.scheme,
                    verify_certs=es_conf.verify_certs,
                    timeout=es_conf.timeout, 
                    max_retries=es_conf.max_retries, 
                    retry_on_timeout=es_conf.retry_on_timeout, 
                )
                
                return RagcarAsyncElasticSemanticSearch(
                    self.config, 
                    encoder=encoder, 
                    db=es
                )
            else:
                from elasticsearch import Elasticsearch
            
                es = Elasticsearch(
                    es_conf.host_n, 
                    http_auth=es_conf.http_auth, 
                    scheme=es_conf.scheme,
                    verify_certs=es_conf.verify_certs,
                    timeout=es_conf.timeout, 
                    max_retries=es_conf.max_retries, 
                    retry_on_timeout=es_conf.retry_on_timeout, 
                )
                
                return RagcarElasticSemanticSearch(
                    self.config, 
                    encoder=encoder, 
                    db=es
                )
        
        encoder = RagcarSentenceFactory(
            tool="sentence_embedding",
            src=self.config.src,
            model=self.config.model_info,
        ).load(device)
        
        return RagcarSemanticSearch(
            self.config, 
            encoder=encoder
        )
        

class RagcarSemanticSearch(RagcarBiencoderBase):
    
    def __init__(
        self, 
        config,
        encoder
    ):
        super().__init__(config)
        self._encoder = encoder
    
    def _retrieve_relevant_entries(
        self, 
        query_embedding: torch.tensor, 
        corpus_embeddings: torch.tensor, 
        top_k: int,
        min_score: float
    ):
        """
        Retrieves the most relevant entries from a corpus of embeddings based on a query embedding,
        considering only entries with a score above a minimum threshold.
        
        Args:
            query_embedding (torch.tensor): The embedding of the query to search for.
            corpus_embeddings (torch.tensor): The embeddings of the corpus to search within.
            top_k (int): The number of top relevant results to return.
            min_score (float): The minimum score a result must have to be considered relevant.
        
        Returns:
            list: A list of hits, each a list of dictionaries with 'score' and potentially other metadata,
                filtered by the minimum score threshold.
        """
        query_embedding = query_embedding.cuda()  # Move query embedding to GPU for faster computation
        corpus_embeddings = corpus_embeddings.cuda()  # Move corpus embeddings to GPU
        
        hits = util.semantic_search(query_embedding, corpus_embeddings, top_k=top_k)  # Perform semantic search
        
        if min_score > 0:  # Filter hits by minimum score if specified
            hits = [[entry for entry in hit if entry['score'] >= min_score] for hit in hits]
        
        return hits
    
    def predict(
        self,
        query,
        corpus,
        **kwargs,
    ) -> list:
        """
        Predicts the most relevant entries from a corpus based on a query using sentence embeddings.
        
        Args:
            query (str, list, or torch.Tensor): The query to be matched. Can be a single sentence, a list of sentences,
                                                or a precomputed tensor of embeddings.
            corpus (str, list, or torch.Tensor): The corpus to be searched. Can be a single sentence, a list of sentences,
                                                a list of paragraphs/documents, or a precomputed tensor of embeddings.
            top_k (int, optional): The number of top matching entries to retrieve. Defaults to 5.
            min_score (float, optional): The minimum score threshold for an entry to be considered relevant. Defaults to 0.
        
        Returns:
            List[dict]: A list of dictionaries representing the relevant entries. Each entry contains metadata such as
                        the similarity score.
        """
        top_k = kwargs.get("top_k", 5)
        min_score = kwargs.get("min_score", 0)
        
        if isinstance(query, torch.Tensor):
            query_embeddings = query
        else:
            query_embeddings = self._encoder.predict(query, convert_to_tensor=True)
        
        if isinstance(corpus, torch.Tensor):
            corpus_embeddings = corpus
        else:
            corpus_embeddings = self._encoder.predict(corpus, convert_to_tensor=True)
        
        outputs = self._retrieve_relevant_entries(query_embeddings, corpus_embeddings, top_k, min_score)
        
        return outputs


class RagcarElasticSemanticSearch(RagcarBiencoderBase):
    
    def __init__(
        self, 
        config,
        encoder,
        db,
    ):
        super().__init__(config)
        self._encoder = encoder
        self._client = db
    
    def _get_embedding_from_es(self, doc_id: str, vector_field:str, index_n:str) -> Optional[torch.Tensor]:
        """
        Retrieves the embedding vector from Elasticsearch for a given document ID and vector field.

        Args:
            doc_id (str): The unique identifier of the document in the Elasticsearch index.
            vector_field (str): The name of the field in the document that contains the embedding vector.
            index_n (str): The name of the Elasticsearch index where the document is stored.
        
        Returns:
            torch.Tensor: The embedding vector as a PyTorch tensor, or None if the document or vector field does not exist.
        
        Raises:
            ValueError: If no document is found for the given ID, if the document does not have the specified vector field,
                        or if an error occurs during the retrieval process.
        """
        try:
            response = self._client.search(index=index_n, body={"query": {"term": {"_id": doc_id}}})
            hits = response['hits']['hits']
            
            if not hits:
                raise ValueError(f"No documents found for id {doc_id} in index {self.index_n}")
            
            result = hits[0]['_source']
            vector_value = result.get(vector_field)
            
            if vector_value is None:
                raise ValueError(f"No vector found for field {vector_field} in document {doc_id}")
            
            return torch.Tensor(vector_value)
        except Exception as e:
            raise ValueError(f"Error occurred while retrieving document {doc_id} from Elasticsearch: {str(e)}")
    
    def _retrieve_relevant_entries_from_es(
        self, 
        query_embedding: torch.Tensor,
        vector_field: str,
        index_n: str,
        knn: bool,
        top_k: int,
        min_score: float,
        source_fields: list,
        filter: list,
        must_not: list
    ) -> list:
        """
        Retrieves relevant documents from Elasticsearch based on cosine similarity between the query embedding and document vectors using either k-NN or script score query.
        
        Args:
            query_embedding (torch.Tensor): The query embedding vector.
            vector_field (str): The name of the document field that contains the embedding vector to compare against.
            index_n (str): The name of the Elasticsearch index to search in.
            knn (bool): If True, use k-NN search. If False, use script score query.
            top_k (int): The number of top documents to retrieve.
            min_score (float): The minimum score documents must have to be included in the results.
            source_fields (list): The list of source fields to include in the returned documents.
            filter (list): The filter criteria for the query.
            must_not (list): The criteria that documents must not match.
        
        Returns:
            list: A list of documents that match the query criteria, each as a dictionary, sorted by their relevance based on cosine similarity.
        """
        body = {
            "from": 0,
            "size": top_k,
            "min_score": min_score,
            "_source": source_fields,
        }
        
        if knn:
            body["query"] = {
                "bool": {
                    "filter": filter,
                    "must_not": must_not,
                    "must": [
                        {
                            "elastiknn_nearest_neighbors": {
                                "field": vector_field,
                                "model": "lsh",
                                "similarity": "cosine",
                                "candidates": top_k*10,
                                "vec": {
                                    "values": query_embedding.cpu().numpy()
                                }
                            }
                        }
                    ]
                }
            }
        else:
            body["query"] = {
                "script_score": {
                    "query" : {
                        "bool": {
                            "filter": filter,
                            "must_not": must_not,
                        }
                    },
                    "script": {
                        "source": f"cosineSimilarity(params.query_vector, '{vector_field}')+1.0",
                        "params": {
                            "query_vector": query_embedding.cpu().numpy()
                        }
                    }
                }
            }
        
        response = self._client.search(index=index_n, body=body, request_timeout=5)
        return response['hits']['hits']
    
    def predict(
        self,
        query,
        corpus,
        **kwargs,
    ) -> list:
        """
        Conducts sentence embedding and retrieves relevant entries from Elasticsearch based on the provided query.
        
        Args:
            query (str or list or Tensor): The input query or list of queries, or their embeddings.
            corpus (str): The target corpus or the name of the Elasticsearch dense_vector field.
            index_n (str): The name of the Elasticsearch index to search.
            top_k (int, optional): The number of top matching entries to retrieve (default: 5).
            min_score (float, optional): The minimum similarity score for entries to be considered relevant (default: 0).
            doc_id (bool, optional): If True, treats the input query as an Elasticsearch document ID for direct retrieval (default: False).
            knn (bool, optional): If True, uses the k-nearest neighbors algorithm for retrieval (default: False).
            source_fields (list, optional): Specifies which fields to include in the returned documents (default: []).
            filter (list, optional): Specifies filter criteria for the search, each filter as a dictionary (default: []).
            must_not (list, optional): Specifies conditions for documents to be excluded, each condition as a dictionary (default: []).

        Returns:
            List[dict]: A list of dictionaries, each representing a relevant entry retrieved from Elasticsearch.
        """
        index_n = kwargs.get("index_n")
        top_k = kwargs.get("top_k", 5)
        min_score = kwargs.get("min_score", 0) + 1.0
        doc_id = kwargs.get("doc_id", False)
        knn = kwargs.get("knn", False)
        source_fields = kwargs.get("source_fields", [])
        filter = [{"exists": {"field": corpus}}]
        filter.extend(kwargs.get("filter", []))
            
        if doc_id:
            if isinstance(query, str):
                query_embeddings = self._get_embedding_from_es(query, corpus, index_n)
            elif isinstance(query, list):
                query_embeddings = torch.stack([self._get_embedding_from_es(q, corpus, index_n) for q in query])
        else:
            if isinstance(query, torch.Tensor):
                query_embeddings = query
            else:
                query_embeddings = self._encoder.predict(query, convert_to_tensor=True)
        
        if query_embeddings.dim() > 1:
            outputs = []
            for i, query_embedding in enumerate(query_embeddings):
                must_not = kwargs.get("must_not", [])
                
                if doc_id:
                    must_not.append({"match": {"_id": query[i]}})
                
                outputs.append(
                    self._retrieve_relevant_entries_from_es(
                        query_embedding=query_embedding,
                        vector_field=corpus,
                        index_n=index_n,
                        knn=knn,
                        top_k=top_k,
                        min_score=min_score,
                        source_fields=source_fields,
                        filter=filter,
                        must_not=must_not
                    )
                )
        else:
            must_not = kwargs.get("must_not", [])

            if doc_id:
                must_not.append({"match": {"_id": query}})
                
            outputs = self._retrieve_relevant_entries_from_es(
                query_embedding=query_embeddings,
                vector_field=corpus,
                index_n=index_n,
                knn=knn,
                top_k=top_k,
                min_score=min_score,
                source_fields=source_fields,
                filter=filter,
                must_not=must_not
            )
        
        return outputs


class RagcarAsyncElasticSemanticSearch(RagcarAsyncBiencoderBase):
    
    def __init__(
        self, 
        config,
        encoder,
        db,
    ):
        super().__init__(config)
        self._encoder = encoder
        self._client = db
    
    async def _get_embedding_from_es(self, doc_id: str, vector_field: str, index_n: str) -> Optional[torch.Tensor]:
        """
        Asynchronously retrieves the embedding vector from Elasticsearch for a given document ID and vector field.

        Args:
            doc_id (str): The unique identifier of the document in the Elasticsearch index.
            vector_field (str): The name of the field in the document that contains the embedding vector.
            index_n (str): The name of the Elasticsearch index where the document is stored.
        
        Returns:
            Optional[torch.Tensor]: The embedding vector as a PyTorch tensor, or None if the document or vector field does not exist.
        
        Raises:
            ValueError: If no document is found for the given ID, if the document does not have the specified vector field,
                        or if an error occurs during the retrieval process.
        """
        try:
            response = await self._client.search(index=index_n, body={"query": {"term": {"_id": doc_id}}})
            hits = response['hits']['hits']
            
            if not hits:
                raise ValueError(f"No documents found for id {doc_id} in index {index_n}")
            
            result = hits[0]['_source']
            vector_value = result.get(vector_field)
            
            if vector_value is None:
                raise ValueError(f"No vector found for field {vector_field} in document {doc_id}")
            
            return torch.Tensor(vector_value)
        except Exception as e:
            raise ValueError(f"Error occurred while retrieving document {doc_id} from Elasticsearch: {str(e)}")
    
    async def _retrieve_relevant_entries_from_es(
        self, 
        query_embedding: torch.Tensor,
        vector_field: str,
        index_n: str,
        knn: bool,
        top_k: int,
        min_score: float,
        source_fields: list,
        filter: list,
        must_not: list
    ) -> list:
        """
        Asynchronously retrieves relevant documents from Elasticsearch based on cosine similarity between the query embedding and document vectors using either k-NN or script score query.
        
        Args:
            query_embedding (torch.Tensor): The query embedding vector.
            vector_field (str): The name of the document field that contains the embedding vector to compare against.
            index_n (str): The name of the Elasticsearch index to search in.
            knn (bool): If True, use k-NN search. If False, use script score query.
            top_k (int): The number of top documents to retrieve.
            min_score (float): The minimum score documents must have to be included in the results.
            source_fields (list): The list of source fields to include in the returned documents.
            filter (list): The filter criteria for the query.
            must_not (list): The criteria that documents must not match.
        
        Returns:
            list: A list of documents that match the query criteria, each as a dictionary, sorted by their relevance based on cosine similarity.
        """
        body = {
            "from": 0,
            "size": top_k,
            "min_score": min_score,
            "_source": source_fields,
        }
        
        if knn:
            body["query"] = {
                "bool": {
                    "filter": filter,
                    "must_not": must_not,
                    "must": [
                        {
                            "elastiknn_nearest_neighbors": {
                                "field": vector_field,
                                "model": "lsh",
                                "similarity": "cosine",
                                "candidates": top_k*10,
                                "vec": {
                                    "values": query_embedding.cpu().numpy()
                                }
                            }
                        }
                    ]
                }
            }
        else:
            body["query"] = {
                "script_score": {
                    "query" : {
                        "bool": {
                            "filter": filter,
                            "must_not": must_not,
                        }
                    },
                    "script": {
                        "source": f"cosineSimilarity(params.query_vector, '{vector_field}')+1.0",
                        "params": {
                            "query_vector": query_embedding.cpu().numpy()
                        }
                    }
                }
            }
        
        response = await self._client.search(index=index_n, body=body, request_timeout=5)
        return response['hits']['hits']
    
    async def predict(
        self,
        query,
        corpus,
        **kwargs,
    ) -> list:
        """
        Asynchronously conducts sentence embedding and retrieves relevant entries from Elasticsearch based on the provided query.
        
        Args:
            query (str or list or Tensor): The input query or list of queries, or their embeddings.
            corpus (str): The target corpus or the name of the Elasticsearch dense_vector field.
            index_n (str): The name of the Elasticsearch index to search.
            top_k (int, optional): The number of top matching entries to retrieve (default: 5).
            min_score (float, optional): The minimum similarity score for entries to be considered relevant (default: 0).
            doc_id (bool, optional): If True, treats the input query as an Elasticsearch document ID for direct retrieval (default: False).
            knn (bool, optional): If True, uses the k-nearest neighbors algorithm for retrieval (default: False).
            source_fields (list, optional): Specifies which fields to include in the returned documents (default: []).
            filter (list, optional): Specifies filter criteria for the search, each filter as a dictionary (default: []).
            must_not (list, optional): Specifies conditions for documents to be excluded, each condition as a dictionary (default: []).

        Returns:
            List[dict]: A list of dictionaries, each representing a relevant entry retrieved from Elasticsearch.
        """
        index_n = kwargs.get("index_n")
        top_k = kwargs.get("top_k", 5)
        min_score = kwargs.get("min_score", 0)
        min_score = min_score + 1.0
        doc_id = kwargs.get("doc_id", False)
        knn = kwargs.get("knn", False)
        source_fields = kwargs.get("source_fields", [])
        filter = [{"exists": {"field": corpus}}]
        filter.extend(kwargs.get("filter", []))
            
        if doc_id:
            if isinstance(query, str):
                query_embeddings = await self._get_embedding_from_es(query, corpus, index_n)
            elif isinstance(query, list):
                tasks = [self._get_embedding_from_es(q, corpus, index_n) for q in query]
                query_embeddings = torch.stack(await asyncio.gather(*tasks))
        else:
            if isinstance(query, torch.Tensor):
                query_embeddings = query
            else:
                query_embeddings = self._encoder.predict(query, convert_to_tensor=True)
        
        if query_embeddings.dim() > 1:
            tasks = []
            for i, query_embedding in enumerate(query_embeddings):
                must_not = kwargs.get("must_not", [])
                
                if doc_id:
                    must_not.append({"match": {"_id": query[i]}})
                
                tasks.append(
                    self._retrieve_relevant_entries_from_es(
                        query_embedding=query_embedding,
                        vector_field=corpus,
                        index_n=index_n,
                        knn=knn,
                        top_k=top_k,
                        min_score=min_score,
                        source_fields=source_fields,
                        filter=filter,
                        must_not=must_not
                    )
                )
            
            outputs = await asyncio.gather(*tasks)
        else:
            must_not = kwargs.get("must_not", [])

            if doc_id:
                must_not.append({"match": {"_id": query}})
                
            outputs = await self._retrieve_relevant_entries_from_es(
                query_embedding=query_embeddings,
                vector_field=corpus,
                index_n=index_n,
                knn=knn,
                top_k=top_k,
                min_score=min_score,
                source_fields=source_fields,
                filter=filter,
                must_not=must_not
            )
        
        return outputs


import logging
import uuid
import time
from functools import reduce
from typing import Optional, Union, Any, Dict, List

from tqdm import tqdm
import numpy as np
from elasticsearch import Elasticsearch, helpers
from elasticsearch.exceptions import ConnectionTimeout, ConnectionError, NotFoundError

from ragcar.tools.utils.base import RagcarToolBase


logger = logging.getLogger(__name__)


class ElasticsearchVectorIndexer:
    """Class for indexing sentence embeddings in Elasticsearch using Sentence Transformers."""

    def __init__(
        self, 
        host: str, 
        http_auth=None, 
        scheme="http", 
        verify_certs=False, 
        timeout=5, 
        max_retries=2, 
        retry_on_timeout=True
    ):
        """
        Initializes an Elasticsearch client with specified configurations.

        Args:
            host (str): The host address of the Elasticsearch server.
            http_auth (optional): Authentication credentials (username, password) if required.
            scheme (str, optional): The protocol scheme, 'http' or 'https'. Defaults to 'http'.
            verify_certs (bool, optional): Whether to verify SSL certificates. Defaults to False.
            timeout (int, optional): Timeout for performing the operations in seconds. Defaults to 5.
            max_retries (int, optional): Maximum number of retries before an operation fails. Defaults to 2.
            retry_on_timeout (bool, optional): Whether to retry on timeout. Defaults to True.

        Raises:
            ValueError: If an invalid host is provided.
            ConnectionError: If unable to connect to the Elasticsearch instance.
        """
        self.client = Elasticsearch(
            host,
            http_auth=http_auth,
            scheme=scheme,
            verify_certs=verify_certs,
            timeout=timeout,
            max_retries=max_retries,
            retry_on_timeout=retry_on_timeout
        )
        
        if self.client.ping():
            logger.info("Connected to Elasticsearch.")
        else:
            logger.error("Could not connect to Elasticsearch.")

    def get_data(
        self, 
        index_name: str, 
        field_to_embed: Union[str, List[str]], 
        query_size: int = 1000
    ) -> List[Dict[str, Any]]:
        """
        Retrieves documents from an Elasticsearch index using the Scroll API for efficient large-scale data retrieval.

        Args:
            index_name (str): The name of the Elasticsearch index to fetch documents from.
            field_to_embed (Union[str, List[str]]): The path to the text field(s) in the documents for embedding purposes.
            query_size (int, optional): The number of documents to retrieve per scroll batch. Defaults to 1000.

        Returns:
            List[Dict[str, Any]]: A list of documents retrieved from Elasticsearch.

        Raises:
            ValueError: If an invalid field path is provided.
        """
        logger.info(f"Fetching documents from index: {index_name} using Scroll API")

        query = {
            "size": query_size,
            "track_total_hits": True, 
            "_source": ".".join(field_to_embed),
            "query": {
                "match_all": {}
            }
        }

        # Initialize the scroll
        res = self.client.search(index=index_name, body=query, scroll='1m')  # '1m' means each scroll request lasts 1 minute
        total = res['hits']['total']['value']
        scroll_id = res['_scroll_id']
        
        total_docs = []

        # Using tqdm for progress bar
        with tqdm(total=total, desc="Fetching from Elasticsearch", unit="docs") as pbar:
            while len(res['hits']['hits']):
                total_docs.extend(res['hits']['hits'])
                pbar.update(len(res['hits']['hits']))  # Update the progress bar
                
                # Scroll to next batch of data
                res = self.client.scroll(scroll_id=scroll_id, scroll='1m')
                scroll_id = res['_scroll_id']

        logger.info(f"Retrieved {len(total_docs)} documents from index: {index_name}")

        return total_docs
    
    def embed_data(
        self, 
        embedder: RagcarToolBase,
        metadata: List[Dict[str, Any]], 
        field_to_embed: Union[str, List[str]] = 'text', 
    ) -> List[List[float]]:
        """
        Generates embeddings for specified text data using the given embedding model.

        Args:
            embedder (RagcarToolBase): The embedding model instance to use for generating embeddings.
            metadata (List[Dict[str, Any]]): A list of dictionaries containing the data for which embeddings will be generated.
            field_to_embed (Union[str, List[str]]): The key(s) in the metadata dict pointing to the text data for embedding.

        Returns:
            List[List[float]]: A list containing the generated embeddings, with each embedding being a list of floats.

        Raises:
            ValueError: If the provided metadata does not match the expected format or if there's an issue with accessing the data to embed.
        """
        sentences = []
        for item in metadata:
            item = item.get('_source', item)
            
            if isinstance(metadata[0], dict):
                if isinstance(field_to_embed, str):
                    sentences.append(item[field_to_embed])
                else:
                    sentences.append(reduce(lambda d, key: d[key], field_to_embed, item))
            else:
                raise ValueError("Provided metadata is not in the expected Elasticsearch format.")
            
        logger.info(f"Generating embeddings for {len(sentences)} sentences")
        embeddings = embedder(sentences)
        logger.info("Embeddings generated successfully")
            
        return embeddings
    
    def _create_index_with_mapping(
        self, 
        index_name: str, 
        vector_field: str , 
        dims: int
    ) -> None:
        """
        Creates an Elasticsearch index with a specific mapping for embedding vectors.

        Args:
            index_name (str): The name of the Elasticsearch index to be created.
            vector_field (str): The name of the field where embedding vectors will be stored.
            dims (int): The dimensions of the embedding vectors.

        Returns:
            None
        """
        logger.info(f"Creating index with mapping for index: {index_name}")
        if not self.client.indices.exists(index=index_name):
            mapping = {
                "mappings": {
                    "properties": {
                        vector_field: {
                            "type": "dense_vector",
                            "dims": dims
                        }
                    }
                }
            }
            self.client.indices.create(index=index_name, body=mapping)
            logger.info(f"Created index: {index_name} successfully")
    
    def backup_index(
        self, 
        index_name: str, 
        backup_name: Optional[str] = None,
        overwrite: bool = True
    ) -> None:
        """
        Creates a backup of an Elasticsearch index by reindexing its data into a new index.

        Args:
            index_name (str): The name of the index to backup.
            backup_name (Optional[str]): The name of the backup index. If not provided, a default name is generated.
            overwrite (bool): Whether to overwrite the backup index if it exists. Defaults to True.

        Returns:
            None

        Raises:
            ConnectionTimeout: If the connection to Elasticsearch times out during the backup process.
            ConnectionError: If there is a failure in connecting to Elasticsearch.
            NotFoundError: If the original index does not exist.
            Exception: For any other unexpected errors during the operation.
        """
        if backup_name is None:
            backup_name = f"{index_name}_backup_{str(uuid.uuid4())[:8]}"

        if not self.client.indices.exists(index=index_name):
            logger.error(f"Index {index_name} does not exist. Cannot create backup.")
            return
        
        if self.client.indices.exists(index=backup_name):
            if overwrite:
                self.client.indices.delete(index=backup_name)
                logger.info(f"Deleted existing index {backup_name}.")
            else:
                logger.warning(f"Index {backup_name} already exists. Use 'overwrite=True' to overwrite the existing index.")
                return

        body = {
            "source": {
                "index": index_name
            },
            "dest": {
                "index": backup_name
            }
        }

        try:
            # Try to execute reindex and get task_id
            reindex_response = self.client.reindex(body=body, wait_for_completion=False)
            task_id = reindex_response["task"]
            
            # Check task completion
            task = self.client.tasks.get(task_id=task_id)
            while not task['completed']:
                time.sleep(5)  # Wait for 5 seconds before checking again
                task = self.client.tasks.get(task_id=task_id)
            
            if task['response']:  # Replace this condition with the appropriate way to check task success
                logger.info(f"Backup for index {index_name} created as {backup_name}")
            else:
                logger.error("Backup task failed.")
        except ConnectionTimeout:
            logger.warning("Connection timed out. The task is still running. Please check manually.")
        except ConnectionError:
            logger.error("Failed to connect to Elasticsearch. Please check your connection.")
        except NotFoundError:
            logger.error(f"Index {index_name} not found.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")

    def restore_from_backup(
        self, 
        backup_name: str, 
        index_name: str, 
        overwrite: bool = True
    ) -> None:
        """
        Restores data from a backup index to a target index in Elasticsearch.

        Args:
            backup_name (str): The name of the backup index from which data will be restored.
            index_name (str): The name of the target index to restore data into.
            overwrite (bool): If True, the target index will be deleted and recreated if it already exists. Defaults to True.

        Returns:
            None

        Raises:
            ConnectionTimeout: If the connection to Elasticsearch times out during the restore operation.
            ConnectionError: If there is a failure in connecting to Elasticsearch.
            NotFoundError: If the specified backup index does not exist.
            Exception: For any other unexpected errors during the operation.
        """
        if not self.client.indices.exists(index=backup_name):
            logger.error(f"Backup index {backup_name} does not exist. Cannot restore data.")
            return

        if self.client.indices.exists(index=index_name):
            if overwrite:
                self.client.indices.delete(index=index_name)
                logger.info(f"Deleted existing index {index_name}.")
            else:
                logger.warning(f"Index {index_name} already exists. Use 'overwrite=True' to overwrite the existing index.")
                return
        
        body = {
            "source": {
                "index": backup_name
            },
            "dest": {
                "index": index_name
            }
        }

        try:
            # Try to execute reindex and get task_id
            reindex_response = self.client.reindex(body=body, wait_for_completion=False)
            task_id = reindex_response["task"]
            
            # Check task completion
            task = self.client.tasks.get(task_id=task_id)
            while not task['completed']:
                time.sleep(5)  # Wait for 5 seconds before checking again
                task = self.client.tasks.get(task_id=task_id)
            
            if task['response']:  # Replace this condition with the appropriate way to check task success
                logger.info(f"Data restored to index {index_name} from backup {backup_name}")
            else:
                logger.error("Restore task failed.")
        except ConnectionTimeout:
            logger.warning("Connection timed out. The task is still running. Please check manually.")
        except ConnectionError:
            logger.error("Failed to connect to Elasticsearch. Please check your connection.")
        except NotFoundError:
            logger.error(f"Index {backup_name} not found.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
    
    def delete_data(self, index_name: str, query: Optional[Dict[str, Any]] = None) -> None:
        """
        Deletes documents from an Elasticsearch index based on a provided query or all documents if no query is provided.

        Args:
            index_name (str): The name of the Elasticsearch index from which documents will be deleted.
            query (Optional[Dict[str, Any]]): A dictionary representing the Elasticsearch query to specify which documents to delete. 
                If None, all documents in the index will be deleted.

        Returns:
            None

        Raises:
            Warning: If the specified index does not exist or no documents matched the query.
        """
        if not self.client.indices.exists(index=index_name):
            logger.warning(f"Index {index_name} does not exist.")
            return

        if query is None:
            # If no query is provided, delete all documents from the index.
            query = {
                "query": {
                    "match_all": {}
                }
            }

        # Use the delete by query API of Elasticsearch
        response = self.client.delete_by_query(index=index_name, body=query)
        
        # Log results
        if response.get('deleted', 0) > 0:
            logger.info(f"Deleted {response['deleted']} documents from index: {index_name}")
        else:
            logger.warning(f"No documents matched the query in index: {index_name}")
    
    def index_data(
        self, 
        index_name: str, 
        metadata: Union[List[str], List[Dict[str, Any]]], 
        embeddings: List[List[float]], 
        decimal_precision: int = 5, 
        text_field: str = 'text', 
        vector_field: str = 'vector', 
        field_to_index: Optional[Union[str, List[str]]] = None
    ) -> None:
        """
        Indexes or updates documents in an Elasticsearch index with provided embeddings and metadata.

        Args:
            index_name (str): Name of the Elasticsearch index to target.
            metadata (Union[List[str], List[Dict[str, Any]]]): Metadata associated with embeddings, can be a list of strings or dictionaries.
            embeddings (List[List[float]]): Numeric vectors representing embeddings to be indexed.
            decimal_precision (int, optional): Number of decimal places to keep for embedding vectors. Defaults to 5.
            text_field (str): Field name for the text data within the metadata. Defaults to 'text'.
            vector_field (str): Field name where embeddings will be stored. Defaults to 'vector'.
            field_to_index (Optional[Union[str, List[str]]], optional): Specific field(s) to update in the document, 
                if None, updates or indexes the whole document.

        Returns:
            None

        Raises:
            ConnectionTimeout: If the connection to Elasticsearch times out during indexing.
            ConnectionError: If there is a failure in connecting to Elasticsearch.
            Exception: For any other unexpected errors during the operation.
        """
        logger.info(f"Indexing data into index: {index_name}")
        actions = []

        # Determine operation type based on index existence
        op_type = "update" if self.client.indices.exists(index=index_name) else "index"
        if op_type == "index":
            self._create_index_with_mapping(index_name, vector_field, len(embeddings[0]))

        for i, item in enumerate(metadata):
            item = item.get('_source', item)
            doc_id = item.get('_id', str(uuid.uuid4()))
            
            action_data = {
                vector_field: np.round(embeddings[i].tolist(), decimal_precision).tolist()
            }
            
            if field_to_index:
                if isinstance(field_to_index, list):
                    for field in field_to_index:
                        action_data[field] = item.get(field)
                else:
                    action_data[field_to_index] = item.get(field_to_index)
            else:
                if isinstance(item, dict):
                    action_data[text_field] = item[text_field]
                else: # If it's just a list of strings
                    action_data[text_field] = item
            
            action = {
                "_op_type": op_type,
                "_index": index_name,
                "_id": doc_id
            }
            
            if op_type == "index":
                action["_source"] = action_data
            else:
                action["doc"] = action_data
                action["doc_as_upsert"] = True
                
            actions.append(action)

        try:
            helpers.bulk(self.client, actions)
            logger.info(f"Indexed data into index: {index_name} successfully")
        except ConnectionTimeout:
            logger.warning("Connection timed out. Some documents may not have been indexed. Please check manually.")
        except ConnectionError:
            logger.error("Failed to connect to Elasticsearch. Please check your connection.")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")


import os
import json
import zipfile
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Optional, Union

import gdown
from dotenv import load_dotenv


@dataclass_json
@dataclass
class TransformerConfig:
    """Configuration for transformer-based model fine-tuned for the STS Task"""
    model_n: str
    token: Optional[str] = None


@dataclass_json
@dataclass
class GoogledriveInfo:
    """Information for a transformer-based model fine-tuned for the STS Task, stored in Google Drive"""
    model_n: str
    model_url: str


@dataclass_json
@dataclass
class OpenaiConfig:
    """Configuration for using GPT-based models with OpenAI"""
    model_n: str
    api_key: str


@dataclass_json
@dataclass
class ClovastudioConfig:
    """Configuration for using HyperCLOVA-based models with NAVER Clova Studio"""
    model_n: str
    api_key: str
    app_key: str


@dataclass
class EncoderInfo:
    """Information for a transformer-based model fine-tuned for the STS Task"""
    src: str
    model: Union[str, dict]


@dataclass_json
@dataclass
class ElasticsearchConfig:
    """Configuration for Elasticsearch as a vector database"""
    encoder_key: EncoderInfo
    host_n: str
    http_auth: Optional[tuple] = None
    scheme: Optional[str] = "http"
    verify_certs: Optional[bool] = True
    timeout: Optional[int] = 5
    max_retries: Optional[int] = 2
    retry_on_timeout: Optional[bool] = True


def get_save_dir(save_dir: str = None) -> str:
    """
    Get a save directory for models or files.

    If a save directory is specified by the user, it returns that directory.
    Otherwise, it defaults to '~/.cache/gdown/models', creating the directory if it does not exist.
    
    Args:
        save_dir(str, optional): User-defined save directory; defaults to None.
    
    Returns:
        str: Set save directory.
    """
    if save_dir:  # If user wants to manually define save directory
        os.makedirs(save_dir, exist_ok=True)
        return save_dir

    # Default save directory: ~/.cache/gdown/models
    default_save_dir = os.path.expanduser('~/.cache/gdown/models')
    os.makedirs(default_save_dir, exist_ok=True)
    
    return default_save_dir

def download_from_url(info: GoogledriveInfo, root_dir: str) -> str:
    """
    Downloads a model from Google Drive and extracts it to the specified directory.

    Args:
        info (GoogledriveInfo): An instance of GoogledriveInfo containing the model's name
            and Google Drive URL. This class provides structured information to identify
            and locate the model to be downloaded.
        root_dir (str): The root directory where the model should be saved. If this directory
            does not exist, it will be created during the download process. The model's contents
            will be extracted into this directory.

    Returns:
        str: The path to the downloaded and extracted model. This path is constructed by 
             combining the root directory and the model's name, providing a direct reference
             to the extracted model's location.
    """
    model_path = os.path.join(root_dir, info.model_n)
    
    if not os.path.exists(model_path):
        gdown.download(url=info.model_url, output="{}.zip".format(info.model_n), quiet=False, fuzzy=True)
        working_dir = os.getcwd()
        source = os.path.join(working_dir, "{}.zip".format(info.model_n))
        source_file = zipfile.ZipFile(source)
        source_file.extractall(root_dir)
        os.remove(source)
        
    return TransformerConfig(model_n=model_path)

def download_or_load(
    src: str,
    model_info: Union[str, dict]
) -> TransformerConfig:
    """
    Downloads or loads a transformer-based model, depending on the source and model information provided.
    
    Args:
        src (str): Specifies the source of the model.
        model_info (Union[str, dict]): Information about the model to be downloaded or loaded. Structure depends on the value of `src`:
            - When `src` is 'model_name_or_path', `model_info` can be:
                - A string specifying the Hugging Face model name or the local path to the model.
                - A dictionary for private Hugging Face models, containing:
                    - 'model_n': The name of the model.
                    - 'token': The Hugging Face access token for private models.
            - When `src` is 'googledrive', `model_info` should be a dictionary with:
                - 'model_n': Name of the model, used for naming the downloaded file.
                - 'model_url': Google Drive URL of the source model file to download.
        save_dir (str, optional): Custom directory path for saving the model. If not provided, a default cache directory is used.
    
    Returns:
        TransformerConfig: An object containing the configuration for the model, including the model name (`model_n`)
                           and, if applicable, the Hugging Face access token (`token`) for private models.
    """
    if src == "model_name_or_path":  # Hugging Face model or local model
        if isinstance(model_info, dict):  # private models
            return TransformerConfig.from_json(json.dumps(model_info))
        
        return TransformerConfig(model_n=model_info)
    
    # Google Drive
    info = GoogledriveInfo.from_json(json.dumps(model_info))
    root_dir = get_save_dir()  # Default save directory
    return download_from_url(info, root_dir)

def get_openai_config(model_info: Union[str, dict]) -> OpenaiConfig:
    """
    Constructs an OpenaiConfig object from given model information or environment variables.

    Args:
        model_info (Union[str, dict]): A dictionary containing the model name ('model_n') and API key ('api_key'),
                                       or a string specifying the model name. If a string is provided, the API key
                                       is fetched from the 'OPENAI_API_KEY' environment variable.

    Returns:
        OpenaiConfig: Configuration object containing the model name and API key for OpenAI API access.
    """
    if isinstance(model_info, dict):
        return OpenaiConfig.from_json(json.dumps(model_info))
    else:
        load_dotenv()
        
        api_key = os.getenv('OPENAI_API_KEY')
        if api_key is None:
            raise ValueError("OPENAI_API_KEY environment variable is not set")
        
        return OpenaiConfig(model_n=model_info, api_key=api_key)

def get_clova_config(model_info: Union[str, dict]) -> ClovastudioConfig:
    """
    Constructs a ClovastudioConfig object from given model information or environment variables.

    Args:
        model_info (Union[str, dict]): A dictionary containing the model name ('model_n'), API key ('api_key'), and application key ('app_key'), 
                                       or a string specifying the model name. The 'api_key' corresponds to the original variable 'X-NCP-APIGW-API-KEY' and 
                                       'app_key' corresponds to 'X-NCP-CLOVASTUDIO-API-KEY'. If a string is provided, the API and app keys are fetched from 
                                       the environment variables, respectively.

    Returns:
        ClovastudioConfig: Configuration object containing the model name, API key, and app key for NAVER Clova Studio API access.
    """
    load_dotenv()
    
    if isinstance(model_info, dict):
        return ClovastudioConfig.from_json(json.dumps(model_info))
    else:
        clovastudio_api_key = os.getenv('X-NCP-APIGW-API-KEY')
        if clovastudio_api_key is None:
            raise ValueError("X-NCP-APIGW-API-KEY environment variable is not set")
        
        clovastudio_app_key = os.getenv('X-NCP-CLOVASTUDIO-API-KEY')
        if clovastudio_app_key is None:
            raise ValueError("X-NCP-CLOVASTUDIO-API-KEY environment variable is not set")
        
        return ClovastudioConfig(
            model_n=model_info, 
            api_key=clovastudio_api_key, 
            app_key=clovastudio_app_key
        )

def get_es_config(host_info: dict) -> ElasticsearchConfig:
    """
    Constructs an ElasticsearchConfig object from a given host information dictionary.

    Args:
        host_info (dict): A dictionary containing the Elasticsearch configuration details, including the encoder key
                          ('encoder_key'), host name ('host_n'), HTTP authentication credentials ('http_auth' - optional),
                          connection scheme ('scheme' - optional), certificate verification ('verify_certs' - optional),
                          connection timeout ('timeout' - optional), maximum retries on failure ('max_retries' - optional),
                          and whether to retry on timeout ('retry_on_timeout' - optional).

    Returns:
        ElasticsearchConfig: Configuration object containing the connection information for Elasticsearch.
    """
    return ElasticsearchConfig.from_json(json.dumps(host_info))


import logging
from typing import Optional, Union, List

import torch
import numpy as np
import asyncio

from ragcar.tools.utils.base import RagcarFactoryBase, RagcarSimpleBase, RagcarAsyncSimpleBase
from ragcar.tools.utils.model_config import download_or_load, get_openai_config, get_clova_config


logger = logging.getLogger(__name__)


class RagcarSentenceFactory(RagcarFactoryBase):
    """
    Sentence embeddings to represent text as dense vectors.
    
    Examples:
        >>> se = Ragcar(tool="sentence_embedding", src="gdrive")
        >>> se("'Ragcar'는 'Retrieval-Augmented Generative Companion for Advanced Research'의 준말입니다.")
        array([-1.25807121e-01, -3.77116412e-01,  7.40933239e-01, -1.15971670e-01, 
        -6.07750058e-01,  8.26456621e-02, -4.65455428e-02, -1.41872734e-01, 
        -2.74574943e-02, -2.61776805e-01,  4.37410325e-01, -6.18349612e-01, 
        ...
        -9.96537209e-02,  3.22387397e-01,  2.22027674e-01, -3.13533843e-01], 
        dtype=float32)  # (1, hidden dim)
        
        >>> se(
        ... "'Ragcar'는 'Retrieval-Augmented Generative Companion for Advanced Research'의 준말입니다.",
        ... convert_to_tensor=True
        ... )
        tensor([-1.2581e-01, -3.7712e-01,  7.4093e-01, -1.1597e-01, -6.0775e-01, 
        8.2646e-02, -4.6546e-02, -1.4187e-01, -2.7457e-02, -2.6178e-01, 
        4.3741e-01, -6.1835e-01,  3.4471e-01,  4.4601e-02, -6.4906e-01, 
        ...
        2.8595e-01,  2.9512e-01, -5.6431e-01,  4.9555e-01, -9.9654e-02, 
        3.2239e-01,  2.2203e-01, -3.1353e-01], device='cuda:0')
    """
    
    def __init__(
        self, 
        tool: str, 
        src: str, 
        model: Optional[Union[str, dict]],
        device: Optional[str] = None,
        use_async: bool = False
    ):
        super().__init__(tool, src, model, device)
        self.use_async = use_async
    
    @staticmethod
    def get_available_srcs():
        return [
            "model_name_or_path",
            "googledrive",
            "openai",
            "clova",
        ]
    
    @staticmethod
    def get_available_models():
        return {
            "model_name_or_path": [
                "leewaay/kpf-bert-base-klueNLI-klueSTS-MSL512",
                "leewaay/klue-roberta-base-klueNLI-klueSTS-MSL512",
                "leewaay/klue-roberta-large-klueNLI-klueSTS-MSL512",
                "MODELS_SUPPORTED(https://huggingface.co/models?pipeline_tag=sentence-similarity)"
            ],
            "googledrive": [
                "YOUR_MODEL"
            ],
            "openai": [
                "text-embedding-3-large",
                "text-embedding-3-small",
                "text-embedding-ada-002",
                "MODELS_SUPPORTED(https://platform.openai.com/docs/models)"
            ],
            "clova": [
                "YOUR_MODEL(https://www.ncloud.com/product/aiService/clovaStudio)"
            ]
        }
    
    def load(self, device: str):
        """
        Loads and initializes the model based on configuration settings.

        Args:
            device (str): The computing device ('cpu' or 'cuda') where the model will be loaded.

        Returns:
            object: The initialized model ready for use.
        """
        if self.config.src == "clova":
            from ragcar.models.clova import ClovaEmbedding
            
            api_conf = get_clova_config(self.config.model_info)
            
            model = ClovaEmbedding(
                api_conf.model_n, 
                api_conf.api_key, 
                api_conf.app_key
            )
            
            if self.use_async:
                return RagcarAsyncGenerativeAISentVec(self.config, model)
            else:
                return RagcarGenerativeAISentVec(self.config, model)
        
        if self.config.src == "openai":
            from ragcar.models.openai import OpenaiEmbedding
            
            api_conf = get_openai_config(self.config.model_info)
            
            model = OpenaiEmbedding(
                api_conf.model_n,
                api_conf.api_key
            )
            
            if self.use_async:
                return RagcarAsyncGenerativeAISentVec(self.config, model)
            else:
                return RagcarGenerativeAISentVec(self.config, model)
        
        from sentence_transformers import SentenceTransformer
        
        model_conf = download_or_load(
            self.config.src, 
            self.config.model_info
        )
        
        model = SentenceTransformer(model_conf.model_n, device)
        
        return RagcarSentenceTransformersSentVec(self.config, model)


class RagcarSentenceTransformersSentVec(RagcarSimpleBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    def predict(
        self, 
        sent: Union[str, List[str]], 
        **kwargs
    ) -> Union[np.ndarray, torch.Tensor]:
        """
        Generates sentence embeddings using the Sentence Transformer model.

        Args:
            sent (Union[str, List[str]]): The sentence or list of sentences to be embedded.
            batch_size (int): Batch size for processing multiple sentences.
            convert_to_tensor (bool): If True, converts the embeddings to PyTorch tensors.

        Returns:
            Union[np.ndarray, torch.Tensor]: The embedding(s) of the input sentence(s) as NumPy arrays or PyTorch tensors.
        """
        batch_size = kwargs.get("batch_size", 32)
        convert_to_tensor = kwargs.get("convert_to_tensor", False)
        
        outputs = self._model.encode(sent, batch_size=batch_size, convert_to_tensor=convert_to_tensor)
        
        return outputs


class RagcarGenerativeAISentVec(RagcarSimpleBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    def predict(
        self, 
        sent: Union[str, List[str]], 
        **kwargs
    ) -> Union[List[float], torch.Tensor]:
        """
        Generates embeddings using a generative AI model for sentences.

        Args:
            sent (Union[str, List[str]]): The sentence or list of sentences for embedding.
            **kwargs: Additional keyword arguments including:
                - batch_size (int): Batch size for processing multiple sentences.
                - convert_to_tensor (bool): If True, converts the embeddings to PyTorch tensors.

        Returns:
            Union[List[float], torch.Tensor]: The embedding(s) of the input sentence(s), either as lists of floats or PyTorch tensors.
        """
        batch_size = kwargs.get("batch_size", 32)
        convert_to_tensor = kwargs.get("convert_to_tensor", False)
        
        if isinstance(sent, str):
            outputs = self._model.encode(sent)
            if convert_to_tensor:
                outputs = torch.tensor(self._model.encode(sent))
        
            return outputs
        
        if isinstance(sent, list):
            outputs = []
            for i in range(0, len(sent), batch_size):
                batch = sent[i:i + batch_size]
                if convert_to_tensor:
                    results = [torch.tensor(self._model.encode(s)) for s in batch]
                else:
                    results = [self._model.encode(s) for s in batch]
                outputs.extend(results)
        
            if convert_to_tensor:
                return torch.stack(outputs)
            else:
                return outputs
        

class RagcarAsyncGenerativeAISentVec(RagcarAsyncSimpleBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    async def predict(
        self, 
        sent: Union[str, List[str]], 
        **kwargs
    ) -> Union[List[float], torch.Tensor]:
        """
        Asynchronously generates embeddings using a generative AI model for sentences.

        Args:
            sent (Union[str, List[str]]): The sentence or list of sentences for embedding.
            **kwargs: Additional keyword arguments including:
                - batch_size (int): Batch size for processing multiple sentences.
                - convert_to_tensor (bool): If True, converts the embeddings to PyTorch tensors.

        Returns:
            Union[List[float], torch.Tensor]: The embedding(s) of the input sentence(s), either as lists of floats or PyTorch tensors, generated asynchronously.
        """
        batch_size = kwargs.get("batch_size", 32)
        convert_to_tensor = kwargs.get("convert_to_tensor", False)
        
        if isinstance(sent, str):
            outputs = await self._model.aencode(sent)
            if convert_to_tensor:
                outputs = torch.tensor(self._model.aencode(sent))
                
            return outputs
        
        if isinstance(sent, list):
            outputs = []
            for i in range(0, len(sent), batch_size):
                batch = sent[i:i + batch_size]
                tasks = [self._model.aencode(s) for s in batch]
                results = await asyncio.gather(*tasks)
                if convert_to_tensor:
                    outputs.extend([torch.tensor(r) for r in results])
                else:
                    outputs.extend(results)
        
            if convert_to_tensor:
                return torch.stack(outputs)
            else:
                return outputs


from typing import Optional, Union, List

import torch

from ragcar.tools.utils.base import RagcarFactoryBase, RagcarSimpleBase, RagcarAsyncSimpleBase
from ragcar.tools.utils.model_config import get_clova_config


class RagcarSegmentationFactory(RagcarFactoryBase):
    """
    Splits text into meaningful segments using CLOVA Studio.
    
    Examples:
        >>> ts = Ragcar(tool="text_segmentation", src="clova", model="https://clovastudio.apigw.ntruss.com/testapp/v1/api-tools/segmentation/{appId}")
        >>> ts("'Ragcar'는 'Retrieval-Augmented Generative Companion for Advanced Research'의 준말입니다.")
        [["'Ragcar'는 'Retrieval-Augmented Generative Companion for Advanced Research'의 준말입니다."]]
    """
    
    def __init__(
        self, 
        tool: str, 
        src: str, 
        model: Optional[Union[str, dict]],
        alpha: int = -100,
        seg_cnt: int = -1,
        post_process: bool = True,
        min_size: int = 300,
        max_size: int = 1000,
        use_async: bool = False
    ):
        super().__init__(tool, src, model)
        self.alpha = alpha
        self.seg_cnt = seg_cnt
        self.post_process = post_process
        self.min_size = min_size
        self.max_size = max_size
        self.use_async = use_async
    
    @staticmethod
    def get_available_srcs():
        return [
            "clova",
        ]
    
    @staticmethod
    def get_available_models():
        return {
            "clova": [
                "YOUR_MODEL(https://www.ncloud.com/product/aiService/clovaStudio)"
            ]
        }
    
    def load(self):
        """
        Loads and initializes the model based on configuration settings.

        Returns:
            object: The initialized model ready for use.
        """
        params = {
            "alpha": self.alpha,
            "seg_cnt": self.seg_cnt,
            "post_process": self.post_process,
            "min_size": self.min_size,
            "max_size": self.max_size
        }
        
        if self.config.src == "clova":
            from ragcar.models.clova import ClovaSegmentation
            
            api_conf = get_clova_config(self.config.model_info)
            
            model = ClovaSegmentation(
                api_conf.model_n, 
                api_conf.api_key, 
                api_conf.app_key
            )
            
            if self.use_async:
                return RagcarAsyncGenerativeAISegmentation(self.config, model, params)
            else:
                return RagcarGenerativeAISegmentation(self.config, model, params)


class RagcarGenerativeAISegmentation(RagcarSimpleBase):
    
    def __init__(self, config, model, params):
        super().__init__(config)
        self._model = model
        self._params = params
    
    def predict(
        self, 
        sent: Union[str, List[str]], 
        **kwargs
    ) -> List[List[str]]:
        """
        Segments text into meaningful units.

        Args:
            sent (str or List[str]): The input text or list of texts to be segmented.
            **kwargs: Additional keyword arguments for segmentation parameters.

        Returns:
            List[List[str]]: A list of segments for each input text.
        """
        outputs = self._model.segment(sent, **self._params)
        return outputs
        

class RagcarAsyncGenerativeAISegmentation(RagcarAsyncSimpleBase):
    
    def __init__(self, config, model, params):
        super().__init__(config)
        self._model = model
        self._params = params
    
    async def predict(
        self, 
        sent: Union[str, List[str]], 
        **kwargs
    ) -> Union[List[float], torch.Tensor]:
        """
        Asynchronously segments text into meaningful units.

        Args:
            sent (str or List[str]): The input text or list of texts to be segmented.
            **kwargs: Additional keyword arguments for segmentation parameters.

        Returns:
            Union[List[float], torch.Tensor]: A list of segments for each input text, or a tensor if specified.
        """
        outputs = await self._model.asegment(sent, **self._params)
        return outputs


import logging
import json
import time
from typing import Optional, Union, Dict, List, Generator, AsyncGenerator

import openai

from ragcar.models.base import OpenaiBase


logger = logging.getLogger(__name__)


class OpenaiEmbedding(OpenaiBase):
    """Class for generating text with OpenAI's Embedding model"""
    
    def __init__(self, model_n: str, api_key: str):
        super().__init__(model_n, api_key)

    def encode(self, text: str) -> List[float]:
        """
        Get embedding for a given text.

        Args:
            text (str): Input text for which embedding is to be generated.

        Returns:
            List[float]: The embedding.
        """
        res, _ = self.fetch(openai.Embedding.create, input=text)
        
        return res['data'][0]['embedding']

    async def aencode(self, text: str) -> List[float]:
        """
        Asynchronously get embedding for a given text.

        Args:
            text (str): Input text for which embedding is to be generated.

        Returns:
            List[float]: The embedding.
        """
        res, _ = await self.afetch(openai.Embedding.acreate, input=text)
        
        return res['data'][0]['embedding']


class OpenaiChatCompletion(OpenaiBase):
    """Class for generating text with OpenAI's Chat model."""
    
    def __init__(
        self, 
        model_n: str, 
        api_key: str, 
        stream: bool, 
        formatting: bool
    ):
        super().__init__(model_n, api_key)
        
        self.stream = stream
        self.formatting = formatting
        
        if self.stream and self.formatting:
            from ragcar.tools.tokenization import RagcarTokenizationFactory
            
            self.tokenizer = RagcarTokenizationFactory(
                tool="tokenization",
                src="openai",
                model=model_n
            ).load()

    def _calculate_tokens(self, prompt: Optional[list] = None, completion: Optional[str] = None):
        if prompt:
            fromatted_prompt = ["""{{"{0}": "{1}"}}""".format(item['role'], item['content']) for item in prompt]
            prompt_str = f"[{', '.join(fromatted_prompt)}]"
            return len(self.tokenizer(str(prompt_str))) - 1
        
        if completion:
            return len(self.tokenizer(completion))
    
    def _get_params(
        self, 
        messages: list, 
        max_tokens: int, 
        temperature: float, 
        top_p: float, 
        frequency_penalty: float, 
        presence_penalty: float, 
        stop_before: Optional[list] = None, 
        functions: Optional[list] = None,
        stream: Optional[bool] = None
    ) -> Dict[str, Union[str, int, float]]:
        params = {
            "messages": messages,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "stop": stop_before,
            "functions": functions,
            "stream": stream
        }
        
        return {k: v for k, v in params.items() if v is not None}
    
    def _output(
        self,
        res,
        request_id,
        start_time
    ) -> Dict[str, Union[str, int, float]]:
        if self.formatting:
            response_time = time.time() - start_time
            formatted_data = self.format_response(res, request_id, response_time)
        
            logger.info(json.dumps(res, ensure_ascii=False, indent=4))
            return formatted_data
        else:
            return res
    
    def _stream_output(
        self,
        res,
        request_id,
        start_time,
        has_functions,
        prompt
    ) -> Generator[str, None, None]:
        message = ""
        for chunk in res:
            if self.formatting:
                logger.info(json.dumps(chunk, ensure_ascii=False, indent=4))
                
                choice = chunk.get('choices', [{}])[0]
                
                delta = choice.get('delta', {})
                finish_reason = choice.get('finish_reason')
                
                if has_functions:
                    function_call = delta.get('function_call', {})
                    content = function_call.get('arguments', None)
                else:
                    content = delta.get('content', None)
                    
                if content:
                    yield {
                        "id": request_id,
                        "event": "chunk",
                        "content": content
                    }
                    
                    message += content
                    
                if finish_reason:
                    response_time = time.time() - start_time
                    
                    prompt_tokens = self._calculate_tokens(prompt=prompt)
                    completion_tokens = self._calculate_tokens(completion=message)
                    
                    default_format = {
                        "choices": [{
                            "message": {
                                "role": "assistant",
                                "content": message,
                            },
                            "finish_reason": finish_reason
                        }],
                        "usage": {
                            "prompt_tokens": prompt_tokens,
                            "completion_tokens": completion_tokens,
                            "total_tokens": prompt_tokens + completion_tokens
                        }
                    }
                    
                    yield {
                        "id": request_id,
                        "event": "finish",
                        "data": self.format_response(default_format, request_id, response_time)
                    }
            else:
                chunk["id"] = request_id
                yield chunk
    
    def create(
        self, 
        messages, 
        **kwargs
    ) -> Union[str, Dict[str, str], Generator[str, None, None]]:
        params = self._get_params(messages, **kwargs)
        
        start_time = time.time()
        res, request_id = self.fetch(openai.ChatCompletion.create, **params)
        
        if params.get('stream'):
            return self._stream_output(res, request_id, start_time, params.get("functions"), messages)
        
        return self._output(res, request_id, start_time)

    async def _astream_output(
        self,
        res,
        request_id,
        start_time,
        has_functions,
        prompt
    ) -> AsyncGenerator[str, None]:
        message = ""
        async for chunk in res:
            if self.formatting:
                logger.info(json.dumps(chunk, ensure_ascii=False, indent=4))
                
                choice = chunk.get('choices', [{}])[0]
                
                delta = choice.get('delta', {})
                finish_reason = choice.get('finish_reason')
                
                if has_functions:
                    function_call = delta.get('function_call', {})
                    content = function_call.get('arguments', None)
                else:
                    content = delta.get('content', None)
                    
                if content is not None:
                    yield {
                        "id": request_id,
                        "event": "chunk",
                        "content": content
                    }
                    
                    message += content
                    
                if finish_reason:
                    response_time = time.time() - start_time
                    
                    prompt_tokens = self._calculate_tokens(prompt=prompt)
                    completion_tokens = self._calculate_tokens(completion=message)
                    
                    default_format = {
                        "choices": [{
                            "message": {
                                "role": "assistant",
                                "content": message,
                            },
                            "finish_reason": finish_reason
                        }],
                        "usage": {
                            "prompt_tokens": prompt_tokens,
                            "completion_tokens": completion_tokens,
                            "total_tokens": prompt_tokens + completion_tokens
                        }
                    }
                    
                    yield {
                        "id": request_id,
                        "event": "finish",
                        "data": self.format_response(default_format, request_id, response_time)
                    }
            else:
                chunk["id"] = request_id
                yield chunk
    
    async def acreate(self, messages, **kwargs) -> Union[str, AsyncGenerator[str, None]]:
        params = self._get_params(messages, **kwargs)
        
        start_time = time.time()
        res, request_id = await self.afetch(openai.ChatCompletion.acreate, **params)
        
        if params.get('stream'):
            return self._astream_output(res, request_id, start_time, params.get("functions"), messages)
        
        return self._output(res, request_id, start_time)
    

class OpenaiCompletion(OpenaiBase):
    """Class for generating text with OpenAI's Completion model"""
    
    def __init__(
        self, 
        model_n: str, 
        api_key: str, 
        stream: bool, 
        formatting: bool
    ):
        super().__init__(model_n, api_key)
        
        self.stream = stream
        self.formatting = formatting
        
        if self.stream and self.formatting:
            from ragcar.tools.tokenization import ToolvaTokenizationFactory
            
            self.tokenizer = ToolvaTokenizationFactory(
                tool="tokenization",
                src="openai",
                model=model_n
            ).load()

    def _calculate_tokens(self, prompt: Optional[list] = None, completion: Optional[str] = None):
        if prompt:
            fromatted_prompt = ["""{{"{0}": "{1}"}}""".format(item['role'], item['content']) for item in prompt]
            prompt_str = f"[{', '.join(fromatted_prompt)}]"
            return len(self.tokenizer(str(prompt_str))) - 1
        
        if completion:
            return len(self.tokenizer(completion))
    
    def _get_params(
        self, 
        prompt: str, 
        max_tokens: int, 
        temperature: float, 
        top_p: float, 
        frequency_penalty: float, 
        presence_penalty: float,
        stop_before: Optional[list] = None,
        stream: Optional[bool] = False
    ) -> Dict[str, Union[str, int, float]]:
        params = {
            "prompt": prompt,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "stop": stop_before,
            "stream": stream
        }
        
        return {k: v for k, v in params.items() if v is not None}
    
    def _output(
        self,
        res,
        request_id,
        start_time
    ) -> Dict[str, Union[str, int, float]]:
        if self.formatting:
            response_time = time.time() - start_time
            formatted_data = self.format_response(res, request_id, response_time)
        
            logger.info(json.dumps(res, ensure_ascii=False, indent=4))
            return formatted_data
        else:
            return res
    
    def _stream_output(
        self, 
        res, 
        request_id, 
        start_time,
        prompt
    ) -> Generator[str, None, None]:
        message = ""
        for chunk in res:
            if self.formatting:
                logger.info(json.dumps(chunk, ensure_ascii=False, indent=4))
                
                choice = chunk.get('choices', [{}])[0]
                
                content = choice[0].get('text', {})
                finish_reason = choice.get('finish_reason')
                
                if content:
                    yield {
                        "id": request_id,
                        "event": "chunk",
                        "content": content
                    }
                    
                    message += content
                
                if finish_reason:
                    response_time = time.time() - start_time
                    
                    prompt_tokens = self._calculate_tokens(prompt=prompt)
                    completion_tokens = self._calculate_tokens(completion=message)
                    
                    default_format = {
                        "choices": [{
                            "text": message,
                            "finish_reason": finish_reason
                        }],
                        "usage": {
                            "prompt_tokens": prompt_tokens,
                            "completion_tokens": completion_tokens,
                            "total_tokens": prompt_tokens + completion_tokens
                        }
                    }
                    
                    yield {
                        "id": request_id,
                        "event": "finish",
                        "data": self.format_response(default_format, request_id, response_time)
                    }
            else:
                chunk["id"] = request_id
                yield chunk
    
    def create(self, prompt, **kwargs) -> Union[str, Generator[str, None, None]]:
        params = self._get_params(prompt, **kwargs)
        
        start_time = time.time()
        res, request_id = self.fetch(openai.Completion.create, **params)
        
        if params.get('stream'):
            return self._stream_output(res, request_id, start_time, params.get("functions"), prompt)
        
        return self._output(res, request_id, start_time)

    async def _astream_output(self, res, request_id) -> AsyncGenerator[str, None]:
        message = ""
        async for chunk in res:
            if self.formatting:
                logger.info(json.dumps(chunk, ensure_ascii=False, indent=4))
                
                choice = chunk.get('choices', [{}])[0]
                
                content = choice[0].get('text', {})
                finish_reason = choice.get('finish_reason')
                
                if content:
                    yield {
                        "id": request_id,
                        "event": "chunk",
                        "content": content
                    }
                    
                    message += content
                
                if finish_reason:
                    response_time = time.time() - start_time
                    
                    prompt_tokens = self._calculate_tokens(prompt=prompt)
                    completion_tokens = self._calculate_tokens(completion=message)
                    
                    default_format = {
                        "choices": [{
                            "text": message,
                            "finish_reason": finish_reason
                        }],
                        "usage": {
                            "prompt_tokens": prompt_tokens,
                            "completion_tokens": completion_tokens,
                            "total_tokens": prompt_tokens + completion_tokens
                        }
                    }
                    
                    yield {
                        "id": request_id,
                        "event": "finish",
                        "data": self.format_response(default_format, request_id, response_time)
                    }
            else:
                chunk["id"] = request_id
                yield chunk
    
    async def acreate(self, prompt, **kwargs) -> Union[str, AsyncGenerator[str, None]]:
        params = self._get_params(prompt, **kwargs)
        
        start_time = time.time()
        res, request_id = await self.afetch(openai.Completion.acreate, **params)
        response_time = time.time() - start_time
        
        if params.get('stream'):
            return self._stream_output(res, request_id, start_time, params.get("functions"), prompt)
        
        return self._output(res, request_id, start_time)


from ragcar.utils.prompt_template import PromptTemplate
from ragcar.utils.token_limiter import TokenLimiter
from ragcar.utils.memory_manager import MemoryManager
from ragcar.utils.chat_summarizer import ChatSummarizer
from ragcar.utils.history_manager import HistoryManager
from ragcar.utils.vector_indexer import ElasticsearchVectorIndexer


from typing import Optional

from ragcar.utils.token_limiter import TokenLimiter
from ragcar.utils.prompt_template import PromptTemplate


class MemoryManager:
    """Class for managing a conversation memory."""
    
    def __init__(
        self, 
        turn_size: int = 1,
        token_limiter: Optional[TokenLimiter] = None
    ):
        """
        Initializes the memory manager with a specified turn size and an optional token limiter.

        Args:
            turn_size (int, optional): The size of the interaction between the human and AI, representing the number of exchanges (turns) to remember. Defaults to 1.
            token_limiter (Optional[TokenLimiter]): An instance of the TokenLimiter class to apply token limits on the conversation memory, or None if no limit is applied.
        """
        self.convo = PromptTemplate()
        self.turn_size = turn_size
        
        self.token_limiter = token_limiter
    
    def save_context(self, input: str, output: str) -> PromptTemplate:
        """
        Saves a new interaction (user input and assistant output) to the conversation history, applying turn size and token limit constraints.

        Args:
            input (str): The latest user input to save.
            output (str): The corresponding assistant output to save.

        Returns:
            PromptTemplate: The updated conversation history, trimmed according to the set turn size and token limit.
        """
        self.convo.user(input)
        self.convo.assistant(output)
        
        # retain only the latest 'turn_size' pairs of conversations.
        while len(self.convo.get_prompt()) > self.turn_size * 2:  # *2 because we have pairs (input, output)
            self.convo.remove_message(0)
        
        # apply token limitation if token_limiter is provided
        if self.token_limiter:
            self.convo = self.token_limiter.cutoff(self.convo)
        
        return self.convo


from __future__ import annotations

import json
from pathlib import Path
from typing import Optional, Union, List, Dict


class PromptTemplate:
    
    
    def __init__(
        self, 
        prompt: Optional[Union[str, List[Dict[str, str]]]] = None,
        user_prefix: Optional[str] = None,
        assistant_prefix: Optional[str] = None
    ):
        """
        Initialize a PromptTemplate instance.
        
        Args:
            prompt (Optional[Union[str, List[Dict[str, str]]]]): The initial conversation prompt. This can be:
                - A single string representing the initial message or scenario.
                - A list of dictionaries, where each dictionary represents an individual message with
                  'role' (e.g., 'user' or 'assistant') and 'content' (the message text).
            user_prefix (Optional[str]): A prefix to prepend to user messages, helping to identify them.
            assistant_prefix (Optional[str]): A prefix to prepend to assistant messages, helping to identify them.
        """
        self.prompt = prompt if isinstance(prompt, (str, list)) else []
        self.user_prefix = user_prefix
        self.assistant_prefix = assistant_prefix

    def say_as(self, role: str, message: str) -> PromptTemplate:
        """
        Adds a message with an associated role to the prompt.

        Args:
            role (str): The role of the speaker, indicating who is the source of the message. This could be any descriptive string like 'user', 'assistant', etc.
            message (str): The content of the message being added to the prompt. This is the actual text that the speaker is said to be conveying.

        Returns:
            PromptTemplate: This instance of PromptTemplate is returned to enable method chaining. This allows for the sequential addition of multiple messages 
                to the prompt by calling this method repeatedly.
        """
        self.prompt.append({"role": role, "content": message})
        return self

    def system(self, message: str) -> PromptTemplate:
        """Shortcut for say_as with 'system' as role"""
        return self.say_as("system", message)

    def user(self, message: str) -> PromptTemplate:
        """Shortcut for say_as with 'user' as role"""
        if self.user_prefix:
            return self.say_as("user", f"{self.user_prefix}: {message}")
        else:
            return self.say_as("user", f"{message}")

    def assistant(self, message: str) -> PromptTemplate:
        """Shortcut for say_as with 'assistant' as role"""
        if self.user_prefix:
            return self.say_as("assistant", f"{self.assistant_prefix}: {message}")
        else:
            return self.say_as("assistant", f"{message}")

    def remove_message(self, index: int) -> PromptTemplate:
        """
        Remove a specific message from the prompt by its index.

        Args:
            index (int): The index of the message to remove.

        Returns:
            PromptTemplate: The current instance to allow method chaining.

        Raises:
            IndexError: If the index is out of the range for the prompt list.
        """
        if 0 <= index < len(self.prompt):
            del self.prompt[index]
        else:
            raise IndexError(f"Index {index} is out of range for the prompt list.")
        return self
    
    def get_prompt(self) -> Union[str, List[Dict[str, str]]]:
        """
        Retrieves the current value of the prompt.

        Returns:
            Union[str, List[Dict[str, str]]]: The current value of `prompt`, which may be a single string or a list of dictionaries with string values.
        """
        return self.prompt
    
    def format_text(self, **kwargs: str) -> str:
        """
        Formats the prompt content into a single string using provided formatting parameters.

        Args:
            **kwargs (str): Keyword arguments for string formatting.

        Returns:
            str: A formatted string.
        """
        if isinstance(self.prompt, str):  # If the prompt is a string, format it directly
            return self.prompt.strip().format(**kwargs)
        
        formatted_messages = []
        for message in self.prompt:
            content = message["content"].strip().format(**kwargs)
            if message["role"] == "system":
                formatted_messages.append(f"{content}\n\n")
            else:
                formatted_messages.append(f"{content}\n")

        return ''.join(formatted_messages).strip()

    def format_chat(self, **kwargs: str) -> List[Dict[str, str]]:
        """
        Formats the stored prompt as a list of message dictionaries, optionally using provided string formatting parameters.

        Args:
            **kwargs: Arbitrary keyword arguments used for string formatting of messages. This allows dynamic insertion of content into the prompt messages.

        Returns:
            List[Dict[str, str]]: A formatted list of message dictionaries. Each dictionary contains 'role' and 'content' keys, 
                where 'role' indicates the speaker ('user' or 'assistant') and 'content' is the formatted message text.
        """
        if isinstance(self.prompt, str):  # If the prompt is a string, consider it as a user message
            return [{"role": "user", "content": self.prompt.strip().format(**kwargs)}]
        
        formatted_messages = []
        for message in self.prompt:
            role = message["role"]
            content = message["content"].strip().format(**kwargs)
            formatted_messages.append({"role": role, "content": content})

        # Check if the last message's role is "user" and append the assistant_prefix
        if formatted_messages[-1]["role"] == "user" and self.assistant_prefix:
            formatted_messages[-1]["content"] += f"\n{self.assistant_prefix}: "
        
        # Memory insertion logic
        if "memory" in kwargs and isinstance(kwargs["memory"], PromptTemplate):
            memory_msgs = kwargs["memory"].get_prompt()
            
            memory_position = next(
                (i for i, m in reversed(list(enumerate(formatted_messages))) if m["role"] == "user"),
                None  # Default if no such pattern is found
            )
            
            if memory_position is not None:
                for memory_msg in memory_msgs:
                    memory_content = memory_msg["content"].strip().format(**kwargs)
                    formatted_messages.insert(memory_position, {"role": memory_msg["role"], "content": memory_content})
                    memory_position += 1  # Update position since we've added new messages
        
        return formatted_messages
    
    def save(self, file_path: str) -> None:
        """
        Saves the current prompt template to a specified file in JSON format. This method is useful for persisting conversation templates or configurations.

        Args:
            file_path (str): The file path where the prompt template should be saved. The file must have a '.json' extension.

        Raises:
            ValueError: If the provided file path does not have a '.json' extension, indicating that the file format is not supported.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        file_path.parent.mkdir(parents=True, exist_ok=True)

        if file_path.suffix == ".json":
            with file_path.open("w", encoding="utf-8") as f:
                json.dump(self.prompt, f, indent=4, ensure_ascii=False)
        else:
            raise ValueError(f"{file_path} must be a .json file")

    @classmethod
    def load(cls, file_path: str) -> PromptTemplate:
        """
        Loads a prompt template from a specified JSON file and creates a PromptTemplate instance from it.

        Args:
            file_path (str): The file path from which to load the prompt template. The file must have a '.json' extension.

        Returns:
            PromptTemplate: An instance of PromptTemplate initialized with the data loaded from the specified JSON file.

        Raises:
            ValueError: If the provided file path does not have a '.json' extension, indicating that the file format is not supported.
        """
        if isinstance(file_path, str):
            file_path = Path(file_path)

        if file_path.suffix == ".json":
            with file_path.open("r") as f:
                prompt = json.load(f)
        else:
            raise ValueError(f"{file_path} must be a .json file")

        return cls(prompt)


from typing import Optional, Union

from ragcar.tools.utils.base import RagcarFactoryBase, RagcarGenerationBase, RagcarAsyncGenerationBase
from ragcar.tools.utils.model_config import get_openai_config, get_clova_config
from ragcar.utils.prompt_template import PromptTemplate


class RagcarTextGenerationFactory(RagcarFactoryBase):
    """
    Generates text responses based on input prompts.

    Examples:
        >>> generator = Ragcar(
        ... tool="text_generation", 
        ... src="openai", 
        ... model="gpt-3.5-turbo", 
        ... prompt_template=PromptTemplate("{input} 수도는?"), 
        ... max_tokens=50
        ... )
        >>> generator(input="대한민국")
        '대한민국의 수도는 서울입니다.'
        
        >>> generator = Ragcar(
        ... tool="text_generation", 
        ... src="openai", 
        ... model="gpt-3.5-turbo", 
        ... prompt_template=PromptTemplate("{input} 수도는?"), 
        ... max_tokens=50, 
        ... stream=True
        ... )
        >>> response = generator(input="대한민국")
        >>> for i in response:
        ...     print(i)
        대
        한
        민
        국
        의
         수
        도
        는
         서
        울
        입니다
        .
    """
    
    def __init__(
        self, 
        tool: str, 
        src: str, 
        model: Optional[Union[str, dict]],
        prompt_template: PromptTemplate,
        max_tokens: int=1000,
        temperature: float=0.,
        top_p: float=1.,
        frequency_penalty: float=0., 
        presence_penalty: float=2.0,
        stop_before: Optional[list] = None,
        functions: Optional[list] = None,
        stream: Optional[bool] = False, 
        use_async: Optional[bool] = False,
        formatting: Optional[bool] = False
    ):
        super().__init__(tool, src, model)
        self.prompt_template = prompt_template
        self.max_tokens = max_tokens
        self.temperature = 0.1 if src == 'clova' else temperature
        self.top_p = top_p
        self.frequency_penalty = 0.1 if src == 'clova' else frequency_penalty
        self.presence_penalty  = int(presence_penalty) if src == 'clova' else presence_penalty
        self.stop_before = stop_before
        self.functions = functions
        self.stream = stream
        self.use_async = use_async
        self.formatting = formatting
    
    @staticmethod
    def get_available_srcs():
        return [
            "openai",
            "clova",
        ]
    
    @staticmethod
    def get_available_models():
        return {
            "openai": [
                "gpt-4-turbo-preview",
                "gpt-4",
                "gpt-3.5-turbo",
                "MODELS_SUPPORTED(https://platform.openai.com/docs/models)"
            ],
            "clova": [
                "YOUR_MODEL(https://www.ncloud.com/product/aiService/clovaStudio)"
            ]
        }
    
    def load(self):
        """
        Loads and initializes the model based on configuration settings.

        Returns:
            object: The initialized model ready for use.
        """
        params = {
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "stop_before": self.stop_before
        }
        params = {k: v for k, v in params.items() if v is not None}
        
        # OpenAI
        if self.config.src == "openai":
            api_conf = get_openai_config(self.config.model_info)
            
            params["stream"] = self.stream
            
            if "gpt-" in api_conf.model_n:  # >= gpt-3.5
                from ragcar.models.openai import OpenaiChatCompletion
                
                model = OpenaiChatCompletion(
                    api_conf.model_n,
                    api_conf.api_key,
                    self.stream,
                    self.formatting
                )
                
                params["functions"] = self.functions
                
                if self.use_async:
                    return RagcarAsyncChatGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
                else:
                    return RagcarChatGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
            else:  # gpt-3
                from ragcar.models.openai import OpenaiCompletion
                
                model = OpenaiCompletion(
                    api_conf.model_n,
                    api_conf.api_key,
                    self.stream,
                    self.formatting
                )
        
                if self.use_async:
                    return RagcarAsyncTextGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
                else:
                    return RagcarTextGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
        
        # CLOVA
        if self.config.src == "clova":
            api_conf = get_clova_config(self.config.model_info)
            
            if "HCX" in api_conf.model_n:  # HyperCLOVA X
                from ragcar.models.clova import ClovaChatCompletion
                
                model = ClovaChatCompletion(
                    api_conf.model_n, 
                    api_conf.api_key, 
                    api_conf.app_key, 
                    self.stream,
                    self.formatting
                )
                
                if self.use_async:
                    return RagcarAsyncChatGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
                else:
                    return RagcarChatGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
                
            else:  # HyperCLOVA
                from ragcar.models.clova import ClovaCompletion
                
                model = ClovaCompletion(
                    api_conf.model_n, 
                    api_conf.api_key, 
                    api_conf.app_key,
                    self.formatting
                )
                
                if self.use_async:
                    return RagcarAsyncTextGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )
                else:
                    return RagcarTextGeneration(
                        self.config,
                        model,
                        self.prompt_template,
                        params
                    )


class RagcarTextGeneration(RagcarGenerationBase):
    
    def __init__(self, config, model, prompt_template, params):
        super().__init__(config)
        self._model = model
        self._prompt_template = prompt_template
        self._params = params
    
    def predict(self, **kwargs):
        """
        Generates a text response based on a formatted prompt template and model parameters.

        Args:
            **kwargs (dict): Keyword arguments used to fill placeholders in the prompt template.

        Returns:
            dict: Contains details of the response and the generated text.
        """
        formatted_prompt = self._prompt_template.format_text(**kwargs)
        
        outputs = self._model.create(formatted_prompt, **self._params)
        
        return outputs


class RagcarAsyncTextGeneration(RagcarAsyncGenerationBase):
    
    def __init__(self, config, model, prompt_template, params):
        super().__init__(config)
        self._model = model
        self._prompt_template = prompt_template
        self._params = params
    
    async def predict(self, **kwargs):
        """
        Asynchronously generates a text response based on a formatted prompt template and model parameters.

        Args:
            **kwargs (dict): Keyword arguments used to fill placeholders in the prompt template.

        Returns:
            dict: Contains details of the response and the generated text, obtained asynchronously.
        """
        formatted_prompt = self._prompt_template.format_text(**kwargs)
        
        outputs = await self._model.acreate(formatted_prompt, **self._params)
        
        return outputs


class RagcarChatGeneration(RagcarGenerationBase):
    
    def __init__(self, config, model, prompt_template, params):
        super().__init__(config)
        self._model = model
        self._prompt_template = prompt_template
        self._params = params
    
    def predict(self, **kwargs):
        """
        Generates a chat response for given messages using a specified model and prompt template.

        Args:
            **kwargs (dict): Keyword arguments representing the chat messages used to fill placeholders in the prompt template.

        Returns:
            dict: Contains details of the chat response and the generated text.
        """
        formatted_messages = self._prompt_template.format_chat(**kwargs)
        
        outputs = self._model.create(formatted_messages, **self._params)
        
        return outputs


class RagcarAsyncChatGeneration(RagcarAsyncGenerationBase):
    
    def __init__(self, config, model, prompt_template, params):
        super().__init__(config)
        self._model = model
        self._prompt_template = prompt_template
        self._params = params
    
    async def predict(self, **kwargs):
        """
        Asynchronously generates a chat response for given messages using a specified model and prompt template.

        Args:
            **kwargs (dict): Keyword arguments representing the chat messages used to fill placeholders in the prompt template.

        Returns:
            dict: Contains details of the chat response and the generated text, obtained asynchronously.
        """
        formatted_messages = self._prompt_template.format_chat(**kwargs)
        
        outputs = await self._model.acreate(formatted_messages, **self._params)
        
        return outputs


"""
Ragcar tool-specific factory class
"""

from typing import Optional

import torch

from ragcar.tools.utils.base import RagcarToolBase
from ragcar.tools.utils import (
    TransformerConfig, 
    GoogledriveInfo, 
    OpenaiConfig,
    ClovastudioConfig,
    EncoderInfo,
    ElasticsearchConfig
)
from ragcar.tools import(
    RagcarTokenizationFactory,
    RagcarSentenceFactory,
    RagcarStsFactory,
    RagcarSemanticSearchFactory,
    RagcarTextGenerationFactory,
    RagcarSegmentationFactory
)


SUPPORTED_TOOLS = {
    "tokenization": RagcarTokenizationFactory,
    "sentence_embedding": RagcarSentenceFactory,
    "sentence_similarity": RagcarStsFactory,
    "semantic_search": RagcarSemanticSearchFactory,
    "text_generation": RagcarTextGenerationFactory,
    "text_segmentation": RagcarSegmentationFactory,
}

TOOLS_WITHOUT_DEVICE = [
    "tokenization",
    "text_generation",
    "text_segmentation"
]

SRC_ALIASES = {
    "구글드라이브": "googledrive",
    "구글": "googledrive",
    "gdrive": "googledrive",
    "google": "googledrive",
    "gpt": "openai",
    "clovastudio": "clova",
    "hyperclova": "clova",
    "hc": "clova",
    "hcx": "clova",
    "엘라스틱서치": "elasticsearch",
    "es": "elasticsearch",
    "허깅페이스": "huggingface",
    "hf": "huggingface",
    "tt": "tiktoken",
    "키위": "kiwi",
}

CUSTOMIZABLE_SRC = {
    "model_name_or_path": TransformerConfig,
    "googledrive": GoogledriveInfo,
    "openai": OpenaiConfig,
    "clova": ClovastudioConfig,
    "elasticsearch": ElasticsearchConfig,
}


class Ragcar:
    r"""
    This is a generic class that will return one of the tool-specific model classes of the library
    when created with the `__new__()` method
    """
    
    def __new__(
        cls,
        tool: str,
        src: str = "clova",
        model: Optional[str] = None,
        **kwargs,
    ) -> RagcarToolBase:
        
        if tool not in SUPPORTED_TOOLS:
            raise KeyError("Unknown tool {}, available tools are {}".format(
                tool,
                list(SUPPORTED_TOOLS.keys()),
            ))

        src = src.lower()
        src = SRC_ALIASES[src] if src in SRC_ALIASES else src

        if tool in TOOLS_WITHOUT_DEVICE:
            tool_module = SUPPORTED_TOOLS[tool](
                tool,
                src,
                model,
                **kwargs,
            ).load()
        else:
            # Get device information from torch API
            if kwargs.get("device"):
                device = kwargs.get("device")
            else:
                device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

            # Instantiate tool-specific pipeline module, if possible
            tool_module = SUPPORTED_TOOLS[tool](
                tool,
                src,
                model,
                **kwargs,
            ).load(device)

        return tool_module

    @staticmethod
    def available_tools() -> str:
        """
        Returns a string listing the available tools in the Ragcar project.
        
        Returns:
            str: A message stating the available tools.
        """
        return "Available tools are {}".format(list(SUPPORTED_TOOLS.keys()))

    @staticmethod
    def available_models(tool: str) -> str:
        """
        Returns a string listing the names of available models for a specified tool.
        
        Args:
            tool (str): The name of the tool for which available models are requested.
        
        Returns:
            str: A string that provides the available models for the specified tool.
        
        Raises:
            KeyError: If the specified tool is not supported.
        
        Examples:
            >>> Ragcar.available_models("text_generation")
            "Available models for text_generation are ([src]: openai, [model]: gpt-3.5-turbo, MODELS_SUPPORTED(https://platform.openai.com/docs/models)), 
            ([src]: clova, [model]: HCX-002, MODELS_SUPPORTED(https://www.ncloud.com/product/aiService/clovaStudio))"
        """
        if tool not in SUPPORTED_TOOLS:
            raise KeyError(
                "Unknown tool {} ! Please check available models via `available_tools()`"
                .format(tool))

        srcs = SUPPORTED_TOOLS[tool].get_available_models()
        output = f"Available models for {tool} are "
        for src in srcs:
            srcs[src] = list(map(lambda x: 'None' if x is None else x, srcs[src]))
            output += f"([src]: {src}, [model]: {', '.join(srcs[src])}), "
        
        return output[:-2]
    
    @staticmethod
    def available_customizable_src(tool: str) -> str:
        """
        Returns a string that lists customizable sources for a specified tool in the Ragcar project.
        
        Args:
            tool (str): The name of the tool for which customizable sources are requested.
        
        Returns:
            str: A message that lists the available customizable sources for the specified tool.
        
        Raises:
            KeyError: If the specified tool is not supported or does not exist in the `SUPPORTED_TOOLS` dictionary.
        
        Examples:
            >>> Ragcar.available_customizable_src("tokenization")
            "Available customizable src for sentence_embedding are ['googledrive', 'clovax', 'openai']"
        """
        if tool not in SUPPORTED_TOOLS:
            raise KeyError(
                "Unknown tool {} ! Please check available tools via `available_tools()`"
                .format(tool))

        if tool == "tokenization":
            customizable_srcs = ["null"]
        else:
            tool_module = SUPPORTED_TOOLS[tool]
            available_srcs = tool_module.get_available_models().keys()
            customizable_srcs = set(available_srcs).intersection(CUSTOMIZABLE_SRC.keys())

        return "Available customizable src for {} are {}".format(tool, list(customizable_srcs))
    
    @staticmethod
    def available_model_fields(src: str) -> str:
        """
        Returns a string detailing the available fields and their data types for a specified customizable source (src).
        
        Args:
            src (str): The name of the customizable source.
        
        Returns:
            str: A string formatted to list each field and its corresponding data type for the specified source,
                including nested fields for complex types.
        
        Raises:
            KeyError: If the specified source is not supported or does not exist in the `CUSTOMIZABLE_SRC` dictionary.
        
        Examples:
            >>> Ragcar.available_model_fields("clova")
            'Available fields for clova are ([field]: api_key, [type]: str), ([field]: app_key, [type]: str), ([field]: model_n, [type]: str)'
        """
        if src not in CUSTOMIZABLE_SRC:
            raise KeyError(
                "Unknown customizable src {} ! Please check available customizable src via `available_customizable_src()`"
                .format(src))

        dataclass = CUSTOMIZABLE_SRC[src]
        output = f"Available fields for {src} are "
        for field, field_type in dataclass.__annotations__.items():
            if field_type == EncoderInfo:
                encoder_fields = EncoderInfo.__annotations__
                for encoder_field, encoder_field_type in encoder_fields.items():
                    output += f"([field]: {field}.{encoder_field}, [type]: {encoder_field_type.__name__}), "
            else:
                if hasattr(field_type, '__args__'):
                    type_names = ", ".join([t.__name__ for t in field_type.__args__])
                else:
                    type_names = field_type.__name__
                output += f"([field]: {field}, [type]: {type_names}), "

        return output[:-2]


import re
import unicodedata
from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional, Union, Mapping, List

import torch


@dataclass
class ToolConfig:
    tool: str
    src: str
    model_info: Union[str, dict]
    device: Optional[str]


class RagcarToolBase:
    r"""Tool base class that implements basic functions for prediction"""

    def __init__(self, config: ToolConfig):
        self.config = config

    @property
    def model_info(self):
        return self.config.model_info

    @property
    def src(self):
        return self.config.src

    @abstractmethod
    def predict(
        self,
        text: Union[str, List[str]],
        **kwargs,
    ):
        raise NotImplementedError(
            "`predict()` function is not implemented properly!")

    def __call__(self):
        raise NotImplementedError(
            "`call()` function is not implemented properly!")

    def __repr__(self):
        return f"[TOOL]: {self.config.tool.upper()}\n[TOOL]: {self.config.tool.upper()}\n[MODEL]: {self.config.model_info}"
    
    def _normalize(self, text: str):
        """Unicode normalization and whitespace removal (often needed for contexts)"""
        text = unicodedata.normalize("NFKC", text)
        text = re.sub(r"\s+", " ", text).strip()
        return text


class RagcarFactoryBase(object):
    r"""This is a factory base class that construct tool-specific module"""
    
    def __init__(
        self,
        tool: str,
        src: str,
        model: Optional[Union[str, dict]] = None,
        device: Optional[str] = None,
    ):
        self._available_srcs = self.get_available_srcs()
        self._available_models = self.get_available_models()
        self._model2src = {
            m: s for s, ms in self._available_models.items() for m in ms
        }
        
        # Set default src as very first supported src
        assert (
            src in self._available_srcs
        ), f"Following src are supported for this tool: {self._available_srcs}"
        
        if src is None:
            src = self._available_srcs[0]
            
        # Set default model
        if model is None:
            model = self.get_default_model(src)

        self.config = ToolConfig(tool, src, model, device)
            
    @abstractmethod
    def get_available_srcs(self) -> List[str]:
        raise NotImplementedError(
            "`get_available_tools()` is not implemented properly!")

    @abstractmethod
    def get_available_models(self) -> Mapping[str, List[str]]:
        raise NotImplementedError(
            "`get_available_models()` is not implemented properly!")
    
    @abstractmethod
    def get_default_model(self, src: str) -> str:
        return self._available_models[src][0]
    
    @classmethod
    def load(cls) -> RagcarToolBase:
        raise NotImplementedError(
            "Model load function is not implemented properly!")


class RagcarSimpleBase(RagcarToolBase):
    r"""Simple tool base wrapper class"""

    def __call__(self, text: Union[str, List[str]], **kwargs):
        return self.predict(text, **kwargs)


class RagcarAsyncSimpleBase(RagcarToolBase):
    r"""Simple tool base wrapper class"""

    async def __call__(self, text: Union[str, List[str]], **kwargs):
        return await self.predict(text, **kwargs)


class RagcarBiencoderBase(RagcarToolBase):
    r"""Bi-Encoder base wrapper class"""

    def __call__(
        self,
        sent_a: Union[str, List[str], torch.Tensor],
        sent_b: Union[str, List[str], torch.Tensor] = None,
        **kwargs,
    ):
        if isinstance(sent_a, str):
            sent_a = self._normalize(sent_a)
        elif isinstance(sent_a, list):
            sent_a = [self._normalize(t) for t in sent_a]
            
        if isinstance(sent_b, str):
            sent_b = self._normalize(sent_b)
        elif isinstance(sent_b, list):
            sent_b = [self._normalize(t) for t in sent_b]
        
        if isinstance(sent_a, torch.Tensor) and isinstance(sent_b, torch.Tensor):
            assert sent_a.shape[-1] == sent_b.shape[-1], "The last dimension of the tensors must be the same."

        return self.predict(sent_a, sent_b, **kwargs)


class RagcarAsyncBiencoderBase(RagcarToolBase):
    r"""Async Bi-Encoder base wrapper class"""

    async def __call__(
        self,
        sent_a: Union[str, List[str], torch.Tensor],
        sent_b: Union[str, List[str], torch.Tensor] = None,
        **kwargs,
    ):
        if isinstance(sent_a, str):
            sent_a = self._normalize(sent_a)
        elif isinstance(sent_a, list):
            sent_a = [self._normalize(t) for t in sent_a]
            
        if isinstance(sent_b, str):
            sent_b = self._normalize(sent_b)
        elif isinstance(sent_b, list):
            sent_b = [self._normalize(t) for t in sent_b]
        
        if isinstance(sent_a, torch.Tensor) and isinstance(sent_b, torch.Tensor):
            assert sent_a.shape[-1] == sent_b.shape[-1], "The last dimension of the tensors must be the same."

        return await self.predict(sent_a, sent_b, **kwargs)


class RagcarGenerationBase(RagcarToolBase):
    r"""Simple tool base wrapper class"""

    def __call__(self, **kwargs):
        return self.predict(**kwargs)


class RagcarAsyncGenerationBase(RagcarToolBase):
    r"""Simple tool base wrapper class"""

    async def __call__(self, **kwargs):
        return await self.predict(**kwargs)


from typing import Optional, List

from ragcar.utils.token_limiter import TokenLimiter
from ragcar.utils.chat_summarizer import ChatSummarizer


class HistoryManager:
    """Class for managing and summarizing conversation memory to avoid excessive token usage."""
    
    def __init__(
        self, 
        chat_summarizer: ChatSummarizer,
        token_limiter: Optional[TokenLimiter] = None
    ):
        """
        Initializes the history manager with options for conversation summarization and token limitation.

        Args:
            chat_summarizer (ChatSummarizer): A ChatSummarizer instance used for summarizing each turn of the conversation.
            token_limiter (Optional[TokenLimiter]): An optional TokenLimiter instance to enforce token count limits.
        """
        self.history = []
        self.chat_summarizer = chat_summarizer
        
        self.token_limiter = token_limiter
    
    def save_context(self, input: str, output: str) -> List[str]:
        """
        Stores and summarizes the user input and assistant output into the conversation history.

        Args:
            input (str): User's input.
            output (str): Assistant's output.

        Returns:
            List[str]: The updated conversation history with the latest interaction summarized and added.
        """
        summarized = self.chat_summarizer.summarize(input=input, output=output)
        
        self.history.append(summarized)
        
        # apply token limitation if token_limiter is provided
        if self.token_limiter:
            self.history = self.token_limiter.cutoff(self.history)
        
        return self.history
    
    async def asave_context(self, input: str, output: str) -> List[str]:
        """
        Asynchronously stores and summarizes the user input and assistant output into the conversation history.

        Args:
            input (str): User's input.
            output (str): Assistant's output.

        Returns:
            List[str]: The updated conversation history with the latest interaction summarized and added.
        """
        summarized = await self.chat_summarizer.asummarize(input=input, output=output)
        
        self.history.append(summarized)
        
        # apply token limitation if token_limiter is provided
        if self.token_limiter:
            self.history = self.token_limiter.cutoff(self.history)
        
        return self.history


from typing import Union

from ragcar.tools.utils.base import RagcarToolBase
from ragcar.utils.prompt_template import PromptTemplate


class TokenLimiter:
    """Class for limiting the number of tokens in a conversation list to a predefined maximum."""
    
    def __init__(self, tokenizer: RagcarToolBase, max_tokens: int = 1000) -> None:
        """
        Initializes a token limiter instance with a tokenizer derived from RagcarToolBase and a maximum token limit.

        Args:
            tokenizer (RagcarToolBase): An instance of a class derived from RagcarToolBase for tokenization.
            max_tokens (int, optional): The maximum number of tokens allowed. Defaults to 1000.
        """
        self.tokenizer = tokenizer
        self.max_tokens = max_tokens

    def token_counter(self, corpus: str) -> int:
        """
        Count the number of tokens in the input string.

        Args:
            corpus (str): Input string to be tokenized.

        Returns:
            int: Number of tokens in the input string.
        """
        return len(self.tokenizer(corpus))
    
    def cutoff(self, input_data: Union[list, PromptTemplate], direction: str = "front") -> Union[list, PromptTemplate]:
        """
        Trims the input data to ensure the total token count is below the specified maximum limit. This is achieved by
        removing tokens from the start or end of the input data based on the provided direction.

        Args:
            input_data (Union[list, PromptTemplate]): The input data to be trimmed. Can be either a list of tokens or a PromptTemplate object.
            direction (str, optional): The direction from which tokens should be removed. Accepts 'front' for removal from the beginning 
                or 'back' for removal from the end. Defaults to 'front'.

        Returns:
            Union[list, PromptTemplate]: The trimmed input data with a total token count below the maximum limit.

        Raises:
            ValueError: Raised if an invalid direction is specified (not 'front' or 'back').
            ValueError: Raised if the input data is reduced to a size where it becomes empty or loses its meaningful context, 
                indicating that the initial data exceeded the maximum token limit by a significant margin.
        """
        output_data = input_data.copy()
        if isinstance(output_data, PromptTemplate):
            while self.token_counter(str(output_data.get_prompt())) >= self.max_tokens:
                if len(output_data.get_prompt()) <= 2:
                    raise ValueError(
                        "The combined length of the new input exceeds the max tokens allowed. "
                        "Please increase the max_tokens or reduce the length of the input and output."
                    )
                # Remove the oldest conversation pair
                for _ in range(2):
                    output_data.remove_message(0)
        else:
            tokens_str = str(output_data)
            if direction == "front":
                while self.token_counter(tokens_str) >= self.max_tokens and output_data:
                    if len(output_data) <= 1:
                        raise ValueError(
                            "The combined length of the new input exceeds the max tokens allowed. "
                            "Please increase the max_tokens or reduce the length of the input and output."
                        )
                    output_data.pop(0)
                    tokens_str = str(output_data)
            elif direction == "back":
                while self.token_counter(tokens_str) >= self.max_tokens and output_data:
                    if len(output_data) <= 1:
                        raise ValueError(
                            "The combined length of the new input exceeds the max tokens allowed. "
                            "Please increase the max_tokens or reduce the length of the input and output."
                        )
                    output_data.pop()
                    tokens_str = str(output_data)
            else:
                raise ValueError("Invalid direction. Use 'front' or 'back'.")
            
        return output_data


from typing import Optional

from ragcar.tools.utils.base import RagcarToolBase
from ragcar.tools.text_generation import RagcarTextGenerationFactory
from ragcar.utils.prompt_template import PromptTemplate


class ChatSummarizer:
    """Class for summarizing conversations between a user and a chatbot."""
    
    def __init__(self, summarizer: Optional[RagcarToolBase] = None, max_tokens: int=256, use_async: Optional[bool] = False):
        """
        Initializes the ChatSummarizer with an optional custom summarizer or the default configuration.

        Args:
            summarizer (Optional[RagcarToolBase]): An optional custom summarizer instance of RagcarToolBase or derived class.
            max_tokens (int): The maximum number of tokens allowed for the summary. Defaults to 256.
            use_async (Optional[bool]): Enables asynchronous operation mode if True. Defaults to False.
        """
        if summarizer:
            self.summarizer = summarizer
        else:
            summ_template = PromptTemplate()
            summ_template.user(
                (
                    "Given the following conversation between a user and a chatbot, "
                    "please provide only a concise summary in Korean (2-3 sentences) "
                    "using the format 'User는 ..., AI는 ...':\n\nUser: {input}\nAI: {output}"
                )
            )
            
            self.summarizer = RagcarTextGenerationFactory(
                tool="text_generation", 
                src="openai", 
                model="gpt-3.5-turbo", 
                prompt_template=summ_template, 
                max_tokens=max_tokens, 
                temperature=0., 
                top_p=1., 
                frequency_penalty=0., 
                presence_penalty=2.0, 
                async_mode=use_async,
                formatting=True
            ).load()
    
    def summarize(self, input: str, output: str) -> str:
        """
        Generates a summary for a given user input and assistant output.

        Args:
            input (str): The user's input text.
            output (str): The assistant's output text.

        Returns:
            str: A concise summary of the conversation in Korean.
        """
        return self.summarizer(input=input, output=output)
    
    async def asummarize(self, input: str, output: str) -> str:
        """
        Asynchronously generates a summary for a given user input and assistant output.

        Args:
            input (str): The user's input text.
            output (str): The assistant's output text.

        Returns:
            str: A concise summary of the conversation in Korean, generated asynchronously.
        """
        return await self.summarizer(input=input, output=output)


import logging
import json
import uuid
from typing import Optional, Union, Tuple, Generator, AsyncGenerator, Dict

import requests
import aiohttp
import openai


logger = logging.getLogger(__name__)


class ClovaBase:
    """Base class for HyperCLOVA models"""
    
    # Define the cost per token for each model outside the function
    COST_PER_TOKEN = {
        'HCX-003': 0.005,
        'HCX-003-tuning': 0.03,
        'HCX-002': 0.005,
        'HCX-002-tuning': 0.03,
        'LK-D': 0.04,
        'LK-D-tuning': 0.12,
        'LK-C': 0.015,
        'LK-C-tuning': 0.045,
        'LK-B': 0.0025,
        'LK-B-tuning': 0.0075,
    }
    
    def __init__(self, api_url: str, api_key: str, app_key: str, stream: Optional[bool] = False):
        self.api_url = api_url
        
        parts = api_url.split('/')
        if 'tasks' in api_url:  # HyperCLOVA tuning model
            last_two = parts[-2:]

            if last_two[-1] == 'completions':
                last_two[-1] = 'LK-D'
            elif last_two[-1] == 'chat-completions':
                last_two[-1] = 'HCX-003'

            self.model_n = '/'.join(last_two)
        else:
            self.model_n = parts[-1]
        
        self.headers = {
            'Content-Type': 'application/json; charset=utf-8',
            'X-NCP-APIGW-API-KEY': api_key,
            'X-NCP-CLOVASTUDIO-API-KEY': app_key,
            # 'X-NCP-CLOVASTUDIO-REQUEST-ID': request_id
        }
        
        if stream:
            self.headers["Accept"] = "text/event-stream"
        
        self.stream = stream
    
    def _to_camel(self, snake_str: str) -> str:
        """
        Convert a snake_case string to camelCase.
        
        Args:
            snake_str (str): The snake_case string to be converted.

        Returns:
            str: The converted camelCase string.
        """
        components = snake_str.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])
    
    def _get_model_type(self, model_n: str) -> str:
        """
        Extract the model type from Clova engine string.

        Args:
            model_n (str): The model name string from which to extract the model type

        Returns:
            str: The model type extracted from the engine string.

        Raises:
            ValueError: If the engine string format is invalid.
        """
        try:
            return f"{model_n.split('/')[-1]}-tuning" if "/" in model_n else model_n
        except IndexError:
            raise ValueError("Invalid url string format. Unable to extract model type.")

    def _calculate_charge_per_request(self, model_type: str, total_tokens: int) -> Optional[float]:
        """
        Calculate the total cost of a request based on the number of input and output tokens and the engine used.

        Args:
            model_type (str): The type of the model.
            total_tokens (int): The total number of tokens used in a request.

        Returns:
            Optional[float]: The total cost of the request in KRW or None if the model type is not recognized.
        """
        if model_type not in self.COST_PER_TOKEN:
            return None

        cost_per_input_token = self.COST_PER_TOKEN[model_type]
        total_cost_in_krw = total_tokens * cost_per_input_token

        return total_cost_in_krw
    
    def format_response(
        self, 
        response_data,
        request_id: str,
        response_time: float
    ) -> Dict[str, Union[str, int, float, Dict]]:
        """
        Format the response from the API.

        Args:
            response_data (Dict): The raw response from the API, expected to contain result details.
            request_id (str): The unique identifier of the request.
            response_time (float): The time (in seconds) it took to get the response from the API.

        Returns:
            Dict[str, Union[str, int, float, Dict]]: A dictionary containing formatted response details, including:
                - id (str): The request identifier.
                - model (str): The model name used for the request.
                - content (str): The main content message from the response.
                - finish_reason (str): The reason why the operation was finished.
                - input_tokens (int): The number of tokens in the input.
                - output_tokens (int): The number of tokens in the output.
                - total_tokens (int): The total number of tokens processed.
                - predicted_cost (Union[str, float]): The predicted cost of the operation, or "Unknown" if not calculable.
                - response_time (float): The response time of the API call.
                - ai_filter (Dict): A dictionary with AI filter scores, if applicable.
        """
        result = response_data.get('result', {})
        
        content = result.get('message', {}).get('content', '').strip()
        
        finish_reason = result.get('stopReason')
        
        input_tokens = result.get('inputLength')
        output_tokens = result.get('outputLength')
        total_tokens = (input_tokens or 0) + (output_tokens or 0)
        
        model_type = self._get_model_type(self.api_url)
        predicted_cost = self._calculate_charge_per_request(model_type, total_tokens)
        
        ai_filter = result.get('aiFilter')
        if ai_filter:
            ai_filter = {item['name']: int(item['score']) for item in ai_filter}
        
        formatted_data = {
            "id": request_id,
            "model": self.model_n,
            "content": content,
            "finish_reason": finish_reason,
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "total_tokens": total_tokens,
            "predicted_cost": predicted_cost if predicted_cost is not None else "Unknown",
            "response_time": response_time,
            "ai_filter": ai_filter
        }
            
        return formatted_data
    
    def fetch(self, **kwargs: dict) -> dict:
        """
        Send a POST request to the API.

        Args:
            kwargs (dict): The parameters for the request, which are passed as keyword arguments.

        Returns:
            dict: The response from the API.
        
        Raises:
            RuntimeError: If the API does not respond with a '20000' status code.
        """
        request_id = f"clova-{str(uuid.uuid4())}"
        
        # Convert snake_case keys to camelCase
        kwargs = {self._to_camel(k): v for k, v in kwargs.items()}
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "model": self.api_url,
                    "parameters": kwargs
                }, ensure_ascii=False, indent=4
            )
        )
        
        response = requests.post(
            self.api_url,
            json=kwargs, 
            headers=self.headers,
            stream=self.stream
        )
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "message": "Request completed successfully"
                }, ensure_ascii=False, indent=4
            )
        )
        
        if self.stream:
            return response, request_id
        
        parsed_data = response.json()
        
        if parsed_data['status']['code'] == '20000':
            return parsed_data, request_id
        
        raise RuntimeError((f"Request failed with status code: {parsed_data['status']}"))

    async def afetch(self, **kwargs: dict) -> dict:
        """
        Send an asynchronous POST request to the API.

        Args:
            kwargs (dict): The parameters for the request, which are passed as keyword arguments.

        Returns:
            dict: The response from the API.
        
        Raises:
            RuntimeError: If the API does not respond with a '20000' status code.
        """
        request_id = f"clova-{str(uuid.uuid4())}"
        
        # Convert snake_case keys to camelCase
        kwargs = {self._to_camel(k): v for k, v in kwargs.items()}
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "model": self.api_url,
                    "parameters": kwargs
                }, ensure_ascii=False, indent=4
            )
        )
        
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.api_url, 
                data=json.dumps(kwargs), 
                headers=self.headers
            ) as response:
                logger.info(
                    json.dumps(
                        {
                            "id": request_id,
                            "response": "Request completed successfully"
                        }, ensure_ascii=False, indent=4
                    )
                )
                
                if self.stream:
                    return await response.text(), request_id
                
                data = await response.text()
                parsed_data = json.loads(data)
                
                if parsed_data['status']['code'] == '20000':
                    return parsed_data, request_id
                
                raise RuntimeError((f"Request failed with status code: {parsed_data['status']}"))


class OpenaiBase:
    """Base class for OpenAI models"""
    
    # Define the cost per token for each model outside the function
    # https://openai.com/pricing
    COST_PER_TOKEN = {
        'gpt-4-base': {
            'input': 0.03 / 1000,
            'output': 0.06 / 1000
        },
        'gpt-4-large': {
            'input': 0.06 / 1000,
            'output': 0.12 / 1000
        },
        'gpt-4-turbo': {
            'input': 0.01 / 1000,
            'output': 0.03 / 1000
        },
        'gpt-3.5': {
            'input': 0.0010 / 1000,
            'output': 0.0020 / 1000
        },
        'gpt-3.5-instruct': {
            'input': 0.0015 / 1000,
            'output': 0.0020 / 1000
        },
        'gpt-3.5-tuning': {
            'input': 0.0030 / 1000,
            'output': 0.0060 / 1000
        },
        'davinci': {
            'input': 0.0020 / 1000,
            'output': 0.0020 / 1000
        },
        'davinci-tuning': {
            'input': 0.0120 / 1000,
            'output': 0.0120 / 1000
        },
        'babbage': {
            'input': 0.0004 / 1000,
            'output': 0.0004 / 1000
        },
        'babbage-tuning': {
            'input': 0.0016 / 1000,
            'output': 0.0016 / 1000
        }
    }
    
    def __init__(self, model_n: str, api_key: str):
        self.model_n = model_n
        openai.api_key = api_key

    def _get_model_type(self, model_n: str) -> str:
        """
        Extract the model type from OpenAI engine string.

        Args:
            model_n (str): The model name string from which to extract the model type.

        Returns:
            str: The model type extracted from the engine string.

        Raises:
            ValueError: If the engine string format is invalid.
        """
        if model_n.startswith('text-'):
            model_n = model_n.replace('text-', '', 1)

        if 'gpt-3.5' in model_n:
            if 'instruct' in model_n:
                return 'gpt-3.5-instruct'
            elif model_n.startswith('ft:'):
                return 'gpt-3.5-turbo-tuning'
            else:
                return 'gpt-3.5'
        elif 'gpt-4' in model_n:
            if '32k' in model_n:
                return 'gpt-4-large'
            elif '1106' in model_n:
                return 'gpt-4-turbo'
            else:
                return 'gpt-4-base'
        else:
            try:
                if model_n.startswith('ft:'):
                    return model_n.split('-')[0].replace('ft:', '') + "-tuning"
                else:
                    return model_n.split('-')[0]
            except IndexError:
                raise ValueError("Invalid model name string format. Unable to extract model type.")

    def _calculate_charge_per_request(self, model_type: str, input_tokens: int, output_tokens: int) -> Optional[float]:
        """
        Calculate the total cost of a request based on the number of input and output tokens and the engine used.

        Args:
            model_type (str): The type of the model.
            input_tokens (int): The input number of tokens used in a request.
            output_tokens (int): The output number of tokens used in a request.

        Returns:
            Optional[float]: The total cost of the request in KRW or None if the model type is not recognized.
        """
        if model_type not in self.COST_PER_TOKEN:
            return None

        total_cost_in_usd = 0

        if input_tokens:
            cost_per_input_token = self.COST_PER_TOKEN[model_type]['input']
            total_cost_in_usd += input_tokens * cost_per_input_token

        if output_tokens:
            cost_per_output_token = self.COST_PER_TOKEN[model_type]['output']
            total_cost_in_usd += output_tokens * cost_per_output_token

        return total_cost_in_usd

    def format_response(
        self,
        response_data,
        request_id: str,
        response_time: float
    ) -> Dict[str, Union[str, int, float]]:
        """
        Format the response from the API.

        Args:
            response_data (dict): The raw response data from the API.
            request_id (str): The unique identifier for the request.
            response_time (float): The duration it took to receive the response, in seconds.

        Returns:
            Dict[str, Union[str, int, float]]: A dictionary containing the formatted response data. This includes:
                - id (str): The request ID.
                - model (str): The model name used for the request.
                - content (str): The main content returned by the API.
                - finish_reason (str): The reason provided by the API for the request's completion.
                - input_tokens (int): The number of tokens used in the input.
                - output_tokens (int): The number of tokens generated as output.
                - total_tokens (int): The total number of tokens used (input + output).
                - predicted_cost (Union[str, float]): The predicted cost of the operation, or "Unknown" if not calculable.
                - response_time (float): The response time for the API call.
        """
        choice = response_data.get('choices', [{}])[0]
        
        model_type = self._get_model_type(self.model_n)
        
        if model_type.startswith('gpt'):
            message = choice.get('message', {})
            content = message if message.get('function_call') else message.get('content', '').strip()
        else:
            content = choice.get('text', '').strip()
        
        finish_reason = choice.get('finish_reason')
        
        usage = response_data.get('usage', {})
        input_tokens = usage.get('prompt_tokens')
        output_tokens = usage.get('completion_tokens')
        total_tokens = (input_tokens or 0) + (output_tokens or 0)
        
        predicted_cost = self._calculate_charge_per_request(model_type, input_tokens, output_tokens)
        
        formatted_data = {
            "id": request_id,
            "model": self.model_n,
            "content": content,
            "finish_reason": finish_reason,
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "total_tokens": total_tokens,
            "predicted_cost": predicted_cost if predicted_cost is not None else "Unknown",
            "response_time": response_time
        }
            
        return formatted_data

    def fetch(self, create_fn, **kwargs) -> Union[Tuple[dict, str], Tuple[Generator[dict, None, None], str]]:
        """
        Calls an external service and logs the request and response.

        Args:
            create_fn (Callable): The function to call the external service.
            **kwargs: Arbitrary keyword arguments passed to the external service call.

        Returns:
            Union[Tuple[dict, str], Tuple[Generator[dict, None, None], str]]: A tuple containing the response 
                from the external service and the request ID.
        """
        request_id = f"openai-{str(uuid.uuid4())}"
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "model": self.model_n,
                    "parameters": kwargs
                }, ensure_ascii=False, indent=4
            )
        )
        
        response = create_fn(model=self.model_n, **kwargs)
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "message": "Request completed successfully"
                }, ensure_ascii=False, indent=4
            )
        )
        
        return response, request_id
    
    async def afetch(self, create_fn, **kwargs) -> Union[Tuple[dict, str], Tuple[AsyncGenerator[dict, None], str]]:
        """
        Asynchronously calls an external service and logs the request and response.

        Args:
            create_fn (Callable): The function to call the external service.
            **kwargs: Arbitrary keyword arguments passed to the external service call.

        Returns:
            Union[Tuple[dict, str], Tuple[AsyncGenerator[dict, None], str]]: A tuple containing the response 
                from the external service and the request ID.
        """
        request_id = f"openai-{str(uuid.uuid4())}"
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "model": self.model_n,
                    "parameters": kwargs
                }, ensure_ascii=False, indent=4
            )
        )
        
        response = await create_fn(model=self.model_n, **kwargs)
        
        logger.info(
            json.dumps(
                {
                    "id": request_id,
                    "message": "Request completed successfully"
                }, ensure_ascii=False, indent=4
            )
        )
        
        return response, request_id


from abc import abstractmethod
from typing import Optional, Union, List

import torch
from sentence_transformers import util

from ragcar.tools.utils.base import RagcarFactoryBase, RagcarBiencoderBase
from ragcar.tools.utils.model_config import download_or_load


class RagcarStsFactory(RagcarFactoryBase):
    """
    Measure semantic textual similarity between sentences
    
    Examples:
        >>> sts = Ragcar(tool="sentence_similarity", src="model_name_or_path")
        >>> sts(["흡연", "외도"], ["바람피면 죽는다", "담배피면 죽는다", "라이터", "간통"], sorted=True)
        [
            [tensor(0.5466), 1, 3],
            [tensor(0.5263), 0, 1],
            [tensor(0.4776), 0, 2],
            [tensor(0.4251), 0, 3],
            [tensor(0.2208), 1, 2],
            [tensor(0.1471), 0, 0],
            [tensor(0.1330), 1, 1],
            [tensor(0.1021), 1, 0]
        ]
        >>> sts("저는 경기도에 살고 있어요", ["경기도에 거주하고 있습니다", "경기 시작 10분 전 입니다.", "서울 근교에 집이 있어요"])
        tensor([[0.9404, 0.1166, 0.4393]])
        
        >>> sts = Ragcar(tool="sentence_similarity", src="model_name_or_path", model="leewaay/klue-roberta-large-klueSTS-cross", encoder_type="cross")
        >>> sts("저는 경기도에 살고 있어요", ["경기도에 거주하고 있습니다", "경기 시작 10분 전 입니다.", "서울 근교에 집이 있어요"])
        tensor([[0.9115, 0.0061, 0.1801]])
    """
    
    def __init__(
        self, 
        tool: str, 
        src: str, 
        model: Optional[Union[str, dict]],
        device: Optional[str] = None,
        encoder_type: str = 'bi'
    ):
        super().__init__(tool, src, model, device)
        self.encoder_type = encoder_type
    
    @staticmethod
    def get_available_srcs():
        return [
            "model_name_or_path",
            "googledrive",
            "openai",
            "clova",
        ]
    
    @staticmethod
    def get_available_models():
        return {
            "model_name_or_path": [
                "leewaay/kpf-bert-base-klueNLI-klueSTS-MSL512",
                "leewaay/klue-roberta-base-klueNLI-klueSTS-MSL512",
                "leewaay/klue-roberta-large-klueNLI-klueSTS-MSL512",
                "leewaay/kpf-bert-base-klueSTS-cross",
                "leewaay/klue-roberta-large-klueSTS-cross",
                "MODELS_SUPPORTED(https://huggingface.co/models?pipeline_tag=sentence-similarity)"
            ],
            "googledrive": [
                "YOUR_MODEL"
            ],
            "openai": [
                "text-embedding-3-large",
                "text-embedding-3-small",
                "text-embedding-ada-002",
                "MODELS_SUPPORTED(https://platform.openai.com/docs/models)"
            ],
            "clova": [
                "YOUR_MODEL(https://www.ncloud.com/product/aiService/clovaStudio)"
            ]
        }
    
    def load(self, device: str):
        """
        Loads and initializes the model based on configuration settings.

        Args:
            device (str): The computing device ('cpu' or 'cuda') where the model will be loaded.

        Returns:
            object: The initialized model ready for use.
        """
        if any(src in self.config.src for src in ['model_name_or_path', 'googledrive']) and self.encoder_type == "cross":
            model_conf = download_or_load(
                self.config.src, 
                self.config.model_info
            )
            
            from sentence_transformers.cross_encoder import CrossEncoder
        
            model = CrossEncoder(model_conf.model_path)
        
            return RagcarCrossSts(self.config, model)
        
        from ragcar.tools.sentence_embedding import RagcarSentenceFactory
        
        model = RagcarSentenceFactory(
            tool="sentence_embedding",
            src=self.config.src,
            model=self.config.model_info
        ).load(device)
        
        return RagcarSts(self.config, model)


class RagcarStsBase(RagcarBiencoderBase):
    
    def get_sorted_similarity(self, scores: List[int]) -> List[List[int]]:
        """
        Sorts cosine similarity scores between sentence pairs in descending order.

        Args:
            scores (List[int]): A matrix of similarity scores between sentence pairs.

        Returns:
            List[List[int]]: A list of [score, index1, index2] lists sorted by score in descending order, 
                where 'index1' and 'index2' represent the positions of the sentence pairs in the original input list.
        """
        #Find the pairs with the highest cosine similarity scores
        pairs = []
        for i in range(len(scores)):
            for j in range(len(scores[i])):
                pairs.append([scores[i][j], i, j])
        
        #Sort scores in decreasing order
        pairs = sorted(pairs, key=lambda x: x[0], reverse=True)
        
        return pairs


class RagcarSts(RagcarStsBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    def predict(self, sent_a, sent_b, **kwargs):
        """
        Calculates the semantic similarity between two sets of sentences using a Bi-Encoder architecture.

        Args:
            sent_a (str, list, or torch.Tensor): The first sentence or list of sentences.
            sent_b (str, list, torch.Tensor, or None): The second sentence or list of sentences. If None, calculates self-similarity for `sent_a`.
            sorted (bool): If True, returns similarity scores in descending order.
            batch_size (int): Batch size for processing when handling multiple sentences.
            convert_to_tensor (bool): If True, converts the embeddings to PyTorch tensors.

        Returns:
            If `sent_b` is not None and 'sorted' is False, returns a torch.Tensor containing similarity scores.
            If 'sorted' is True, or if `sent_a` is processed for self-similarity, returns a list of lists containing similarity scores and indices from `sent_a` and `sent_b`.

        Raises:
            ValueError: If the inputs are not both strings, lists of strings, or tensors.
        """
        # If sent_a is a string, wrap it in a list
        if isinstance(sent_a, str):
            sent_a = [sent_a]
            
        # If sent_b is a string, wrap it in a list
        if isinstance(sent_b, str):
            sent_b = [sent_b]

        # If inputs are single list
        if isinstance(sent_a, list) and sent_b is None:
            if len(sent_a) < 10000:
                embeddings_a = self._model.predict(sent_a, convert_to_tensor=True)
                embeddings_b = embeddings_a
                cosine_scores = util.cos_sim(embeddings_a, embeddings_b)
                outputs = []
                for i in range(len(cosine_scores)-1):
                    for j in range(i+1, len(cosine_scores)):
                        outputs.append([cosine_scores[i][j], i, j])
            else:
                outputs = util.paraphrase_mining(self._model, sent_a)
                
            return outputs

        # If inputs are list of strings, encode them into embeddings
        if isinstance(sent_a, list) and isinstance(sent_b, list):
            embeddings_a = self._model.predict(sent_a, convert_to_tensor=True)
            embeddings_b = self._model.predict(sent_b, convert_to_tensor=True)
        # If inputs are tensors, just use them directly
        elif isinstance(sent_a, torch.Tensor) and isinstance(sent_b, torch.Tensor):
            embeddings_a = sent_a
            embeddings_b = sent_b
        else:
            raise ValueError("Inputs should be both strings, lists of strings, or tensors.")
            
        scores = util.cos_sim(embeddings_a, embeddings_b)
        
        if "sorted" in kwargs and kwargs["sorted"] is True:
            outputs = self.get_sorted_similarity(scores)
        else:
            outputs = scores

        return outputs
    
    
class RagcarCrossSts(RagcarStsBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    def predict(self, sent_a, sent_b, **kwargs):
        """
        Computes semantic textual similarity scores between two sets of sentences using a cross-encoder model.

        Args:
            sent_a (str or list): The first sentence or list of sentences to be encoded.
            sent_b (str or list): The second sentence or list of sentences to be encoded.
            **kwargs: Arbitrary keyword arguments, including 'sorted' (bool) to specify whether to return the similarity scores in descending order.

        Returns:
            A torch.Tensor of similarity scores if 'sorted' is False. If 'sorted' is True, returns a list of lists containing similarity scores, and indices from `sent_a` and `sent_b`.

        Raises:
            ValueError: If the inputs are not both strings or lists of strings.
        """
        # If sent_a is a string, wrap it in a list
        if isinstance(sent_a, str):
            sent_a = [sent_a]
            
        # If sent_b is a string, wrap it in a list
        if isinstance(sent_b, str):
            sent_b = [sent_b]

        # If inputs are list of strings, encode them into embeddings
        if isinstance(sent_a, list) and isinstance(sent_b, list):
            scores = []
            
            for a in sent_a:
                #Concatenate the query and all passages and predict the scores for the pairs [query, ]
                pairs = [[a, b] for b in sent_b]
                scores.append(self._model.predict(pairs))
            
            scores = torch.Tensor(scores)
        else:
            raise ValueError("Inputs should be both strings, lists of strings, or tensors.")
            
        if "sorted" in kwargs and kwargs["sorted"] is True:
            outputs = self.get_sorted_similarity(scores)
        else:
            outputs = scores

        return outputs


from setuptools import setup, find_packages

setup(
    name='ragcar', 
    version='0.1.0', 
    url='https://github.com/leewaay/ragcar.git', 
    author='Wonseok Lee', 
    author_email='wonsuklee7020@gmail.com', 
    description='RAGCAR: Retrieval-Augmented Generative Companion for Advanced Research', 
    packages=find_packages(), 
    python_requires='>=3.8',
    install_requires=[
        'torch==2.0.1',
        'python-dateutil',
        'dataclasses_json',
        'python-dotenv',
        'tqdm',
        'pandas',
        'gdown',
        'tiktoken',
        'kiwipiepy',
        'elasticsearch==7.13.1',
        'aiohttp',
        'sentence-transformers==2.4.0',
        'openai==0.28.1',
        'pytorch-lightning==1.1.0',
        'fsspec==2021.4.0'
    ],
)


from typing import Optional, Union, List

from ragcar.tools.utils.base import RagcarFactoryBase, RagcarSimpleBase, RagcarAsyncSimpleBase
from ragcar.tools.utils.model_config import get_clova_config


class RagcarTokenizationFactory(RagcarFactoryBase):
    """
    Tokenizing text or counting the number of tokens.
    
    Examples:
        >>> tokenizer = Ragcar(tool="tokenization", src="huggingface")
        >>> tokenizer("나는 동물을 좋아하는 사람이야")
        ['나', '##는', '동물', '##을', '좋아하', '##는', '사람', '##이', '##야']
        
        >>> tokenizer = Ragcar(tool="tokenization", src="tiktoken")
        >>> tokenizer("나는 동물을 좋아하는 사람이야")
        [
            167,
            224,
            246,
            167,
            232,
            242,
            ...
            234,
            35975,
            112,
            168,
            243,
            120
        ]
        
        >>> tokenizer = Ragcar(tool="tokenization", src="kiwi")
        >>> tokenizer("나는 동물을 좋아하는 사람이야")
        [
            Token(form='나', tag='NP', start=0, len=1),
            Token(form='는', tag='JX', start=1, len=1),
            Token(form='동물', tag='NNG', start=3, len=2),
            Token(form='을', tag='JKO', start=5, len=1),
            Token(form='좋아하', tag='VV', start=7, len=3),
            Token(form='는', tag='ETM', start=10, len=1),
            Token(form='사람', tag='NNG', start=12, len=2),
            Token(form='이', tag='VCP', start=14, len=1),
            Token(form='야', tag='EF', start=15, len=1)
        ]
    """
    
    def __init__(
        self, 
        tool: str, 
        src: str, 
        model: Optional[Union[str, dict]],
        use_async: bool = False
    ):
        super().__init__(tool, src, model)
        self.use_async = use_async
    
    @staticmethod
    def get_available_srcs():
        return [
            "huggingface",
            "tiktoken",
            "openai",
            "clova",
            "kiwi",
        ]
    
    @staticmethod
    def get_available_models():
        return {
            "huggingface": [
                "klue/roberta-large",
                "jinmang2/kpfbert",
                "MODELS_SUPPORTED(https://huggingface.co/models?library=transformers)"
            ],
            "tiktoken": [
                "cl100k_base",
                "p50k_base",
                "r50k_base",
                "gpt2",
                "MODELS_SUPPORTED(https://github.com/openai/tiktoken/blob/main/tiktoken/model.py)"
            ],
            "openai": [
                "gpt-4-turbo",
                "gpt-4",
                "gpt-3.5-turbo",
                "MODELS_SUPPORTED(https://platform.openai.com/docs/models)"
            ],
            "clova": [
                "YOUR_MODEL(https://www.ncloud.com/product/aiService/clovaStudio)"
            ],
            "kiwi": [
                None,
                "YOUR_MODEL"
            ],
        }
    
    def load(self):
        """
        Loads and initializes the model based on configuration settings.

        Returns:
            object: The initialized model ready for use.
        """
        if self.config.src == "tiktoken":
            import tiktoken
            
            model = tiktoken.get_encoding(self.config.model_info)
            
            return RagcarTiktokenTokenizer(self.config, model)
        
        if self.config.src == "openai":
            import tiktoken
            
            model = tiktoken.encoding_for_model(self.config.model_info)
            
            return RagcarTiktokenTokenizer(self.config, model)
        
        if self.config.src == "clova":  # HyperCLOVA X
            api_conf = get_clova_config(self.config.model_info)
            
            if "chat-tokenize" in api_conf.model_n:
                from ragcar.models.clova import ClovaChatTokenization
                
                model = ClovaChatTokenization(
                    api_conf.model_n, 
                    api_conf.api_key, 
                    api_conf.app_key
                )
            else:
                from ragcar.models.clova import ClovaTokenization
                
                model = ClovaTokenization(
                    api_conf.model_n, 
                    api_conf.api_key, 
                    api_conf.app_key
                )
                
            if self.use_async:
                return RagcarAsyncTokenizer(self.config, model)
        
        elif self.config.src == "kiwi":
            from kiwipiepy import Kiwi
            
            model = Kiwi()
            
            if self.config.model_info:
                model.load_user_dictionary(self.config.model_info)
        
        else:
            from transformers import AutoTokenizer

            model = AutoTokenizer.from_pretrained(self.config.model_info)
        
        return RagcarTokenizer(self.config, model)


class RagcarTiktokenTokenizer(RagcarSimpleBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    def predict(self, sent: str, **kwargs) -> List[str]:
        """
        Tokenizes the input sentence into a sequence of tokens.

        Args:
            sent (str): The sentence to be tokenized.

        Returns:
            List[str]: A list of tokens derived from the input sentence.
        """
        token_integers = self._model.encode(sent)
        outputs = self._model.decode_tokens_bytes(token_integers)
        return outputs


class RagcarTokenizer(RagcarSimpleBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    def predict(self, sent: str, **kwargs) -> List[str]:
        """
        Tokenizes a sentence into a sequence of strings.

        Args:
            sent (str): The sentence to be tokenized.

        Returns:
            List[str]: A list of tokens as strings.
        """
        outputs = self._model.tokenize(sent)
        return outputs

class RagcarAsyncTokenizer(RagcarAsyncSimpleBase):
    
    def __init__(self, config, model):
        super().__init__(config)
        self._model = model
    
    async def predict(self, sent: str, **kwargs) -> List[str]:
        """
        Asynchronously tokenizes a sentence into a sequence of strings.

        Args:
            sent (str): The sentence to be tokenized.

        Returns:
            List[str]: A list of tokens as strings, obtained asynchronously.
        """
        outputs = await self._model.atokenize(sent)
        return outputs